# Copyright (c) 2019-2020, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import asyncio
import urllib.parse
from concurrent.futures.thread import ThreadPoolExecutor
from pathlib import Path
from typing import Any, AsyncIterator, List, Optional, Dict, cast
import logging

import backoff
from backoff import on_exception
from minio import Minio, ResponseError
from minio.error import NoSuchBucket
from urllib3.exceptions import MaxRetryError

from .utils import ServiceXException, ServiceXNoFilesInCache


class MinioAdaptor:
    # Threadpool on which downloads occur. This is because the current minio library
    # uses blocking http requests, so we can't use asyncio to interleave them.
    _download_executor = ThreadPoolExecutor(max_workers=5)

    def __init__(
        self,
        minio_endpoint: str,
        secured: bool = False,
        access_key: str = "miniouser",
        secretkey: str = "leftfoot1",
    ):
        self._endpoint = minio_endpoint
        self._secured = secured
        self._access_key = access_key
        self._secretkey = secretkey

        self._client = Minio(
            self._endpoint,
            access_key=self._access_key,
            secret_key=self._secretkey,
            secure=self._secured,
        )

    @on_exception(backoff.constant, ResponseError, interval=0.1)
    def get_files(self, request_id) -> List[str]:
        try:
            return [str(f.object_name) for f in self._client.list_objects(request_id)]
        except MaxRetryError as e:
            msg = (
                f"Unable to reach Minio at {e.pool.host}:{e.pool.port}{e.url}. "
                f"Max retries exceeded."
            )
            raise ServiceXException(msg)
        except NoSuchBucket as e:
            raise ServiceXNoFilesInCache(
                f'No files in cache for request "{request_id}"'
            ) from e

    def get_access_url(self, request_id: str, object_name: str) -> str:
        """Given a request ID and the file name in that request id, return a URL
        that can be directly accessed to download the file.

        Args:
            request_id (str): The request id guid (that is the bucket in minio)
            object_name (str): The file (the object in the minio bucket)

        Returns:
            str: A url good for some amount of time to access the bucket.
        """
        return self._client.presigned_get_object(request_id, object_name)

    def get_s3_uri(self, request_id: str, object_name: str) -> str:
        """Given a request ID and a file name in that request, return an S3 URI that can
        be used by a correctly configured S3 client.

        Args:
        :param request_id: The request id guid (that is the bucket in minio)
        :param object_name: The file (the object in the minio bucket)
        :return:
            A formatted S3 URI to that object
        """
        return urllib.parse.urlunsplit(("s3", request_id, object_name, None, None))

    async def download_file(
        self, request_id: str, bucket_fname: str, output_file: Path
    ) -> None:
        """
        Download a single file to a local temp file.

        Arguments:
            minio_client        Open and authenticated minio client
            request_id          The id of the request we are going after
            bucket_fname        The fname of the bucket
            output_file         Filename where we should write this file.

        Notes:
            - Download to a temp file that is renamed at the end so that a partially
              downloaded file is not mistaken as a full one
            - Run with async, despite minio not being async.
        """
        # Make sure the output directory exists
        output_file.parent.mkdir(parents=True, exist_ok=True)

        # We are going to build a temp file, and download it from there.
        def do_copy() -> None:
            temp_file = output_file.parent / f"{output_file.name}.temp"
            try:
                log = logging.getLogger(__name__)
                log.debug(
                    f"Starting download of minio file for request {request_id}, "
                    f"file {bucket_fname}."
                )
                self._client.fget_object(request_id, bucket_fname, str(temp_file))
                log.debug(
                    f"Finished download of minio file for request {request_id}, "
                    f"file {bucket_fname}."
                )
                temp_file.rename(output_file)
            except Exception as e:
                raise ServiceXException(
                    f"Failed to copy minio bucket {bucket_fname} from request "
                    f"{request_id} to {output_file}"
                ) from e

        # If the file exists, we don't need to do anything.
        if output_file.exists():
            return

        # Do the copy, which might take a while, on a separate thread.
        return await asyncio.wrap_future(self._download_executor.submit(do_copy))


class MinioAdaptorFactory:
    """A factor that will return, when asked, the proper minio adaptor to use for a request
    to get files from ServiceX.
    """

    def __init__(self, always_return: Optional[MinioAdaptor] = None):
        """Create the factor with a possible adaptor to always use

        Args:
            always_return (Optional[MinioAdaptor], optional): The adaptor to always use. If none,
            then one will be created by the best other available method. Defaults to None.
        """
        # Get the defaults setup.
        self._always = always_return
        self._config_adaptor = None

    def from_best(
        self, transaction_info: Optional[Dict[str, str]] = None
    ) -> MinioAdaptor:
        """Using the information we have, create the proper Minio Adaptor with the correct
        endpoint and login information. Order of adaptor generation:

        1. The `always_return` option from the ctor
        1. Use info from the `translation_info`
        1. Use the info from the config handed in to the ctor

        Raises:
            Exception: If we do not have enough information to create an adaptor, this is raised.

        Returns:
            MinioAdaptor: The adaptor that can be used to extract the data from minio for this
            request.
        """
        if self._always is not None:
            logging.getLogger(__name__).debug("Using the pre-defined minio_adaptor")
            return self._always
        if transaction_info is not None:
            keys = [
                "minio-endpoint",
                "minio-secured",
                "minio-access-key",
                "minio-secret-key",
            ]
            if all(k in transaction_info for k in keys):
                logging.getLogger(__name__).debug(
                    "Using the request-specific minio_adaptor"
                )
                return MinioAdaptor(
                    transaction_info["minio-endpoint"],
                    bool(transaction_info["minio-secured"]),
                    transaction_info["minio-access-key"],
                    transaction_info["minio-secret-key"],
                )
        raise ServiceXException(
            "Do not know or have enough information to create a Minio "
            f"Login info ({transaction_info})"
        )


async def find_new_bucket_files(
    adaptor: MinioAdaptor, request_id: str, update: AsyncIterator[Any]
) -> AsyncIterator[str]:
    """
    Each time we get something from the async iterator, check to see if
    there are any files present.
    """
    seen = []
    async for _ in update:
        # Sadly, this is blocking, and so may hold things up
        files = cast(List[str], adaptor.get_files(request_id))

        # If there are new files, pass them on
        for f in files:
            if f not in seen:
                seen.append(f)
                yield f



#  Copyright (c) 2022 , IRIS-HEP
#   All rights reserved.
#
#   Redistribution and use in source and binary forms, with or without
#   modification, are permitted provided that the following conditions are met:
#
#   * Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#   * Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#   * Neither the name of the copyright holder nor the names of its
#     contributors may be used to endorse or promote products derived from
#     this software without specific prior written permission.
#
#   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#   SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#   CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#
#
#
import inspect
from base64 import b64encode
from pathlib import Path
from typing import Optional, List, Callable

from servicex import ServiceXDataset


class ServiceXPythonFunction(ServiceXDataset):
    @staticmethod
    def _encode_function(selection_function: Callable):
        return b64encode(inspect.getsource(selection_function).encode("utf-8")).decode(
            "utf-8"
        )

    async def get_data_rootfiles_async(
        self, selection_function: Callable, title: Optional[str] = None
    ) -> List[Path]:
        return await self._file_return(
            self._encode_function(selection_function), "root-file", title
        )

    async def get_data_awkward_async(
        self, selection_function: Callable, title: Optional[str] = None
    ):
        return self._converter.combine_awkward(
            await self._data_return(
                self._encode_function(selection_function),
                lambda f: self._converter.convert_to_awkward(f),
                title,
            )
        )



import logging
from typing import Dict, Optional, Tuple

from confuse.core import ConfigView

from servicex.ConfigSettings import ConfigSettings
from .utils import ServiceXException


class ServiceXConfigAdaptor:
    """Contains the logic to extract all the configuration information needed, as driven from
    things the user has given us and program parameters. This concentrates all the complex config
    logic in on place, and, hopefully, leaves the logic out of all the other code.
    """

    def __init__(self, config: Optional[ConfigView] = None):
        """The config needed for the app.

        Note: The config is held onto and only queired when the information is required.

        Args:
            config (ConfigView): The config information for the app. If null, then we just use the
                                 standard servicex name.
        """
        self._settings = (
            config if config is not None else ConfigSettings("servicex", "servicex")
        )

    @property
    def settings(self) -> ConfigView:
        """Return the config settings.

        Eventually this should not be used other than for testing!

        Returns:
            ConfigSettings: The config settings that gives us a view on everything the user wants
        """
        return self._settings

    def get_default_returned_datatype(self, backend_name: Optional[str]) -> str:
        """Return the default return data type, given the backend is a certian type.

        Args:
            backend_name (Optional[str]): The backend type string (`xaod`, `uproot`, etc)

        Returns:
            str: The backend datatype, like `root` or `parquet`.
        """
        # Check to see if we know about the backend info
        r = self.get_backend_info(backend_name, "return_data")

        if r is None:
            raise ServiceXException(
                "A default default_return_data is missing from config files "
                "- is servicex installed correctly?"
            )

        return r

    def get_backend_info(self, backend_name: Optional[str], key: str) -> Optional[str]:
        """Find an item in the backend info, searching first for the backend
        name/type and then the defaults with a given type.

        Args:
            backend_name (str): Backend name
            key (str): The key for the info we are after

        Returns:
            Optional[str]: Return a string for the info we are after, or return None if we can't
                           find it.
        """
        config = self._get_backend_info(backend_name)
        return config[key] if key in config else None

    def _get_backend_info(self, backend_name: Optional[str]) -> Dict[str, str]:
        """Returns all the info for a backend name/type.

        Search algorithm is non-trivial:
        1. If `backend_name` is not `None`:
           1. Look at the `api_endpoints` for a `name` matching `backend_name`.
           2. Look at the `api_endpoints` for a `type` matching `backend_name`,
              complain that `name` wasn't present.
           3. Fail if nothing matches
        2. If `backend_name` is None:
           1. Use the first end point in the list, and complain.

        Given the above is done, then look at `backend_types` for a matching `type`,
        and for any key found there not already present, add it, and return the dictionary.

        Args:
            backend_name (str): Name or type of the api end point we are going to look up.

        Returns:
            Dict[str, str]: Attributes for this backend's configuration
        """
        # Find a list of all endpoints.
        # It is an error if this is not specified somewhere.
        endpoints = self.settings["api_endpoints"]

        # If we have a good name, look for exact match
        config: Optional[Dict[str, str]] = None
        log = logging.getLogger(__name__)
        if backend_name is not None:
            # Search the list of end points for a matching name
            for ep in endpoints:
                if ep["name"].exists() and ep["name"].as_str_expanded() == backend_name:
                    config = {k: str(ep[k].as_str_expanded()) for k in ep.keys()}
                    break
            if config is None:
                for ep in endpoints:
                    if (
                        ep["type"].exists()
                        and ep["type"].as_str_expanded() == backend_name
                    ):
                        config = {k: str(ep[k].as_str_expanded()) for k in ep.keys()}
                        log.warning(
                            f'Found backend type matching "{backend_name}". Matching by '
                            'type is depreciated. Please switch to using the "name" '
                            "keyword in your servicex.yaml file."
                        )
                        break
            if config is None:
                # They asked for something explicit, we couldn't find it, so we need to bomb.
                # Given how annoying this is, build a useful error message.
                seen_types = [
                    str(ep["type"].as_str_expanded())
                    for ep in endpoints
                    if ep["type"].exists()
                ]
                seen_names = [
                    str(ep["name"].as_str_expanded())
                    for ep in endpoints
                    if ep["name"].exists()
                ]

                raise ServiceXException(
                    f"Unable to find name/type {backend_name} "
                    "in api_endpoints in servicex.yaml configuration file. Saw"
                    f' only names ({", ".join(seen_names)}) and types '
                    f'({", ".join(seen_types)})'
                )

        # Nope - now we are going to have to just use the first one there.
        else:
            for ep in endpoints:
                log.warning(
                    "No backend name/type requested, "
                    f'using {ep["endpoint"].as_str_expanded()} - please be explicit '
                    "in the ServiceXDataset constructor"
                )
                config = {k: str(ep[k].as_str_expanded()) for k in ep.keys()}
                break

        if config is None:
            # If we are here - then... it just isn't going to work!
            raise ServiceXException(
                "Not even a default set of configurations are here! Bad "
                " install of the servicex package!"
            )

        # Now, extract the type and see if we can figure out any defaults from the
        # `backend_types` info.
        type_lookup = config["type"] if "type" in config else backend_name
        if type_lookup is None:
            return config

        backend_defaults = self.settings["backend_types"]
        for bd in backend_defaults:
            if bd["type"].as_str_expanded() == type_lookup:
                for k in bd.keys():
                    if k not in config:
                        config[k] = str(bd[k].as_str_expanded())

        # Finally, a default return type
        if "return_data" not in config:
            if "default_return_data" in self.settings.keys():
                config["return_data"] = str(
                    self.settings["default_return_data"].as_str_expanded()
                )

        return config

    def get_servicex_adaptor_config(
        self, backend_name: Optional[str] = None
    ) -> Tuple[str, Optional[str]]:
        """Return the servicex (endpoint, token) from a given backend configuration.

        Args:
            backend_name (str): The backend name (like `xaod`) which we hopefully can find in the
            configuration file.

        Returns:
            Tuple[str, str]: The tuple of info to create a `ServiceXAdaptor`: end point,
            token (optionally).
        """
        config = self._get_backend_info(backend_name)

        endpoint = config["endpoint"]
        token = config["token"] if "token" in config else None

        # We can default these to "None"
        return (endpoint, token)  # type: ignore



import asyncio
import functools
import logging
import time
from datetime import timedelta
from pathlib import Path
from typing import (
    Any,
    AsyncGenerator,
    AsyncIterator,
    Awaitable,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    Union,
)

import aiohttp
import backoff
import minio
from backoff import on_exception
from make_it_sync import make_sync

from servicex.servicex_config import ServiceXConfigAdaptor

from .cache import Cache
from .data_conversions import DataConverterAdaptor
from .minio_adaptor import MinioAdaptor, MinioAdaptorFactory, find_new_bucket_files
from .servicex_adaptor import (
    ServiceXAdaptor,
    transform_status_stream,
    trap_servicex_failures,
)
from .servicex_utils import _wrap_in_memory_sx_cache
from .servicexabc import ServiceXABC
from .utils import (
    DatasetType,
    ServiceXException,
    ServiceXFailedFileTransform,
    ServiceXFatalTransformException,
    ServiceXNoFilesInCache,
    ServiceXUnknownDataRequestID,
    ServiceXUnknownRequestID,
    StatusUpdateFactory,
    _run_default_wrapper,
    _status_update_wrapper,
    default_client_session,
    get_configured_cache_path,
    log_adaptor,
    on_exception_itr,
    stream_status_updates,
    stream_unique_updates_only,
)

# The allowed file formats.
# You could modify this if you wanted to add new...
g_allowed_formats = ["parquet", "root-file"]


class StreamInfoBase:
    """Contains base information about results that are streamed back from
    ServiceX.
    """

    def __init__(self, file: str):
        self._file = file

    @property
    def file(self) -> str:
        """Returns the ServiceX filename

        This filename is unique in the dataset, and will be the same across different queries
        against the dataset. It can be used as a key to sort results.

        Notes:

        - May contains non-file system characters

        Returns:
            str: servicex filename
        """
        return self._file


class StreamInfoUrl(StreamInfoBase):
    """Contains information about results that are streamed back from ServiceX.
    Used when a URL to access the data directly from ServiceX is requested.
    """

    def __init__(self, file: str, url: str, bucket: str):
        super().__init__(file)
        self._url = url
        self._bucket = bucket

    @property
    def url(self) -> str:
        """URL that can can be used to stream data back from ServiceX.

        Returns:
            str: The URL of the transformed data for this file.
        """
        return self._url

    @property
    def bucket(self) -> str:
        """Returns the buck name - unique and constant accross transformations.
        Can be used to order the results

        Returns:
            str: The bucket name as produced by ServiceX
        """
        return self._bucket


class StreamInfoPath(StreamInfoBase):
    """Contains information about results that are streamed back from ServiceX.
    Used when the user has requested streaming, but copying the file locally first.
    """

    def __init__(self, file: str, path: Path):
        super().__init__(file)
        self._path = path

    @property
    def path(self) -> Path:
        """Path the the local file of ServiceX data that represents this query.

        Returns:
            Path: The path object that points to the data requested.
        """
        return self._path

    @property
    def url(self) -> str:
        """URI to the locally downloaded file.

        Returns:
            str: The URI of the transformed data for this file.
        """
        return self._path.as_uri()


class StreamInfoData(StreamInfoBase):
    """Contains information about results that are streamed back from ServiceX.
    Used when data (`pandas` or `awkward`) is requested. The data is downloaded from
    ServiceX, converted into the requested format, and then streamed to the user in these
    chunks. There is a single chunk per file.
    """

    def __init__(self, file: str, data: Any):
        super().__init__(file)
        self._data = data

    @property
    def data(self) -> Any:
        """The `pandas.DataFrame` or `awkward` array return

        Returns:
            Any: The ServiceX transformed data for this file.
        """
        return self._data


class ServiceXDataset(ServiceXABC):
    """
    Used to access an instance of ServiceX at an end point on the internet. Support convieration
    by configuration object `config_adaptor` or by creating the adaptors defined in the `__init__`
    function.
    """

    def __init__(
        self,
        dataset: DatasetType,
        backend_name: Optional[str] = None,
        codegen: Optional[str] = None,
        image: Optional[str] = None,
        max_workers: int = 20,
        result_destination: str = "object-store",
        servicex_adaptor: Optional[ServiceXAdaptor] = None,
        minio_adaptor: Optional[Union[MinioAdaptor, MinioAdaptorFactory]] = None,
        cache_adaptor: Optional[Cache] = None,
        status_callback_factory: Optional[StatusUpdateFactory] = _run_default_wrapper,
        local_log: Optional[log_adaptor] = None,
        session_generator: Optional[
            Callable[[], Awaitable[aiohttp.ClientSession]]
        ] = None,
        config_adaptor: Optional[ServiceXConfigAdaptor] = None,
        data_convert_adaptor: Optional[DataConverterAdaptor] = None,
        ignore_cache: bool = False,
    ):
        """
        Create and configure a ServiceX object for a dataset.

        Arguments

            dataset                     Name of a dataset from which queries will be selected.
            backend_name                The type of backend. Used only if we need to find an
                                        end-point. If we do not have a `servicex_adaptor` then this
                                        will default to xaod, unless you have any endpoint listed
                                        in your servicex file. It will default to best match there,
                                        or fail if a name has been given.
            codegen                     The type of code generator passed to the backend to
                                        generate the code that powers the requested transform.
            image                       Name of transformer image to use to transform the data. If
                                        left as default, `None`, then the default image for the
                                        ServiceX backend will be used.
            max_workers                 Maximum number of transformers to run simultaneously on
                                        ServiceX.
            result_destination          Where the transformers should write the results.
                                        Defaults to object-store, but could be used to save
                                        results to a posix volume
            servicex_adaptor            Object to control communication with the servicex instance
                                        at a particular ip address with certian login credentials.
                                        Will be configured via the `config_adaptor` by default.
            minio_adaptor               Object to control communication with the minio servicex
                                        instance.
            cache_adaptor               Runs the caching for data and queries that are sent up and
                                        down.
            status_callback_factory     Factory to create a status notification callback for each
                                        query. One is created per query.
            local_log                   Log adaptor for logging.
            session_generator           If you want to control the `ClientSession` object that
                                        is used for callbacks. Otherwise a single one for all
                                        `servicex` queries is used.
            config_adaptor              Control how configuration options are read from the
                                        a configuration file (e.g. servicex.yaml).
            data_convert_adaptor        Manages conversions between root and parquet and `pandas`
                                        and `awkward`, including default settings for expected
                                        datatypes from the backend.
            ignore_cache                Always ignore the cache on any query for this dataset. This
                                        is only meaningful if no cache adaptor is provided.
                                        Defaults to false - the cache is used if possible.

        Notes:

            -  The `status_callback` argument, by default, uses the `tqdm` library to render
               progress bars in a terminal window or a graphic in a Jupyter notebook (with proper
               jupyter extensions installed). If `status_callback` is specified as None, no
               updates will be rendered. A custom callback function can also be specified which
               takes `(total_files, transformed, downloaded, skipped)` as an argument. The
               `total_files` parameter may be `None` until the system knows how many files need to
               be processed (and some files can even be completed before that is known).
            -  The full description of calling parameters is recorded in the local cache, including
               things like `image` name and tag.
        """
        ServiceXABC.__init__(
            self,
            dataset,
            codegen,
            image,
            max_workers,
            result_destination,
            status_callback_factory,
        )

        # Get the local settings
        config = (
            config_adaptor if config_adaptor is not None else ServiceXConfigAdaptor()
        )

        # Establish the cache that will store all our queries
        self._cache = (
            Cache(
                get_configured_cache_path(config.settings),
                ignore_cache,
                analysis_query_key=backend_name
                if backend_name is not None
                else "default",
            )
            if cache_adaptor is None
            else cache_adaptor
        )

        if not servicex_adaptor:
            # Given servicex adaptor is none, this should be ok. Fixes type checkers
            end_point, token = config.get_servicex_adaptor_config(backend_name)
            servicex_adaptor = ServiceXAdaptor(end_point, token)
        self._servicex_adaptor = servicex_adaptor

        if not minio_adaptor:
            self._minio_adaptor = MinioAdaptorFactory()
        else:
            if isinstance(minio_adaptor, MinioAdaptor):
                self._minio_adaptor = MinioAdaptorFactory(always_return=minio_adaptor)
            else:
                self._minio_adaptor = minio_adaptor

        self._log = log_adaptor() if local_log is None else local_log

        self._session_generator = (
            session_generator
            if session_generator is not None
            else default_client_session
        )

        self._return_types = [config.get_default_returned_datatype(backend_name)]
        self._converter = (
            data_convert_adaptor
            if data_convert_adaptor is not None
            else DataConverterAdaptor(self._return_types[0])
        )

    def first_supported_datatype(
        self, datatypes: Union[List[str], str]
    ) -> Optional[str]:
        """Return the first datatype format that this dataset/servicex instance can return.

        Different instances of `ServiceX` are capable of returning different datatypes. Pass in
        the datatypes that your app supports, and this will return the first one that the servicex
        backend can return.

        Args:
            datatypes (Union[List[str], str]): A single or list of datatypes that are supported by
                                               your app.

        Returns:
            str: The first datatype that is supported. If none of them are, then `None` is
            returned.
        """
        datatypes = [datatypes] if isinstance(datatypes, str) else datatypes
        for dt in datatypes:
            if dt in self._return_types:
                return dt

        return None

    def ignore_cache(self):
        """Return a context manager that, as long as it is held, will cause any queries against
        just this dataset to ignore any locally cached data.

        Returns:
            ContextManager: As long as this is held, the local query cache will be ignored.
        """
        return self._cache.ignore_cache()

    @functools.wraps(ServiceXABC.get_data_rootfiles_async, updated=())
    @_wrap_in_memory_sx_cache
    async def get_data_rootfiles_async(
        self, selection_query: str, title: Optional[str] = None
    ) -> List[Path]:
        return await self._file_return(selection_query, "root-file", title)

    async def get_data_rootfiles_stream(
        self, selection_query: str, title: Optional[str] = None
    ) -> AsyncIterator[StreamInfoPath]:
        """Returns, as an async iterator, each completed batch of work from Servicex.
        The `StreamInfoPath` contains a path where downstream consumers can directly
        access the data.

        Args:
            selection_query (str): The `qastle` query for the data to retreive.

        Yields:
            AsyncIterator[StreamInfoPath]: As ServiceX completes the data, and it is downloaded
                                           to the local machine, the async iterator returns
                                           a `StreamInfoPath` which can be used to access the
                                           file locally.
        """
        async for f_info in self._stream_local_files(
            selection_query, title, "root-file"
        ):  # type: ignore
            yield f_info

    @functools.wraps(ServiceXABC.get_data_parquet_async, updated=())
    @_wrap_in_memory_sx_cache
    async def get_data_parquet_async(
        self, selection_query: str, title: Optional[str] = None
    ) -> List[Path]:
        return await self._file_return(selection_query, "parquet", title)

    async def get_data_parquet_stream(
        self, selection_query: str, title: Optional[str] = None
    ) -> AsyncIterator[StreamInfoPath]:
        """Returns, as an async iterator, each completed batch of work from Servicex.
        The `StreamInfoPath` contains a path where downstream consumers can directly
        access the data.

        Args:
            selection_query (str): The `qastle` query for the data to retreive.

        Yields:
            AsyncIterator[StreamInfoPath]: As ServiceX completes the data, and it is downloaded
                                           to the local machine, the async iterator returns
                                           a `StreamInfoPath` which can be used to access the
                                           file locally.
        """
        async for f_info in self._stream_local_files(
            selection_query, title, "parquet"
        ):  # type: ignore
            yield f_info

    @functools.wraps(ServiceXABC.get_data_pandas_df_async, updated=())
    @_wrap_in_memory_sx_cache
    async def get_data_pandas_df_async(
        self, selection_query: str, title: Optional[str] = None
    ):
        return self._converter.combine_pandas(
            await self._data_return(
                selection_query, lambda f: self._converter.convert_to_pandas(f), title
            )
        )

    @functools.wraps(ServiceXABC.get_data_awkward_async, updated=())
    @_wrap_in_memory_sx_cache
    async def get_data_awkward_async(
        self, selection_query: str, title: Optional[str] = None
    ):
        return self._converter.combine_awkward(
            await self._data_return(
                selection_query, lambda f: self._converter.convert_to_awkward(f), title
            )
        )

    async def get_data_awkward_stream(
        self, selection_query: str, title: Optional[str] = None
    ) -> AsyncGenerator[StreamInfoData, None]:
        """Returns, as an async iterator, each completed batch of work from Servicex
        as a separate `awkward` array. The data is returned in a `StreamInfoData` object.

        Args:
            selection_query (str): The `qastle` query for the data to retreive.

        Yields:
            AsyncIterator[StreamInfoData]: As ServiceX completes the data, and it is downloaded
                                           to the local machine, the async iterator returns
                                           a `StreamInfoData` which can be used to access the
                                           data that has been loaded from the file.
        """
        async for a in self._stream_return(
            selection_query, title, lambda f: self._converter.convert_to_awkward(f)
        ):
            yield a

    async def get_data_pandas_stream(
        self, selection_query: str, title: Optional[str] = None
    ) -> AsyncGenerator[StreamInfoData, None]:
        """Returns, as an async iterator, each completed batch of work from Servicex
        as a separate `pandas.DataFrame` array. The data is returned in a `StreamInfoData` object.

        Args:
            selection_query (str): The `qastle` query for the data to retreive.

        Yields:
            AsyncIterator[StreamInfoData]: As ServiceX completes the data, and it is downloaded
                                           to the local machine, the async iterator returns
                                           a `StreamInfoData` which can be used to access the
                                           data that has been loaded from the file.
        """
        async for a in self._stream_return(
            selection_query, title, lambda f: self._converter.convert_to_pandas(f)
        ):
            yield a

    async def get_data_rootfiles_uri_stream(
        self,
        selection_query: str,
        title: Optional[str] = None,
        as_signed_url: Optional[bool] = False,
    ) -> AsyncIterator[StreamInfoUrl]:
        """Returns, as an async iterator, each completed batch of work from ServiceX.
        The data that comes back includes a `url` that can be accessed to download the
        data.

        Args:
            selection_query (str): The ServiceX Selection
            title (str): Optional title for transform request
            as_signed_url (bool): Return the uri as a presigned http url?
        """
        async for f_info in self._stream_url_buckets(
            selection_query, "root-file", title, as_signed_url
        ):  # type: ignore
            yield f_info

    async def get_data_rootfiles_uri_async(
        self,
        selection_query: str,
        title: Optional[str] = None,
        as_signed_url: Optional[bool] = False,
    ) -> List[StreamInfoUrl]:
        """Returns a list of StreamInfoUrl entries containing a `url` for each output file
        from the transform, taken from get_data_rootfiles_uri_stream.
        The data that comes back includes a `url` that can be accessed to download the
        data.

        Args:
            selection_query (str): The ServiceX Selection
            title (str): Optional title for transform request
            as_signed_url (bool): Return the uri as a presigned http url?
        """
        return [
            f
            async for f in self.get_data_rootfiles_uri_stream(
                selection_query, title=title, as_signed_url=as_signed_url
            )
        ]

    async def get_data_parquet_uri_stream(
        self,
        selection_query: str,
        title: Optional[str] = None,
        as_signed_url: Optional[bool] = False,
    ) -> AsyncIterator[StreamInfoUrl]:
        """Returns, as an async iterator, each of the files from the minio bucket,
        as the files are added there.

        Args:
            selection_query (str): The ServiceX Selection
            title (str): Optional title for transform request
            as_signed_url (bool): Return the uri as a presigned http url?
        """
        async for f_info in self._stream_url_buckets(
            selection_query, "parquet", title, as_signed_url
        ):  # type: ignore
            yield f_info

    async def get_data_parquet_uri_async(
        self,
        selection_query: str,
        title: Optional[str] = None,
        as_signed_url: Optional[bool] = False,
    ) -> List[StreamInfoUrl]:
        """Returns a list of each of the files from the minio bucket.

        Args:
            selection_query (str): The ServiceX Selection
            title (str): Optional title for transform request
            as_signed_url (bool): Return the uri as a presigned http url?
        """
        return [
            f_info
            async for f_info in self.get_data_parquet_uri_stream(
                selection_query, title, as_signed_url
            )
        ]

    async def _file_return(
        self, selection_query: str, data_format: str, title: Optional[str]
    ):
        """
        Given a query, return the list of files, in a unique order, that hold
        the data for the query.

        For certian types of exceptions, the queries will be repeated. For example,
        if `ServiceX` indicates that it was restarted in the middle of the query, then
        the query will be re-submitted.

        Arguments:

            selection_query     `qastle` data that makes up the selection request.
            data_format         The file-based data format (root or parquet)
            title               The title assigned to this transform request.

        Returns:

            data                Data converted to the "proper" format, depending
                                on the converter call.
        """

        async def convert_to_file(f: Path) -> Path:
            return f

        return await self._data_return(
            selection_query, convert_to_file, title, data_format
        )

    @on_exception_itr(
        backoff.constant, ServiceXUnknownRequestID, interval=0.1, max_tries=3
    )
    @on_exception_itr(
        backoff.constant,
        (ServiceXUnknownDataRequestID, minio.error.NoSuchBucket),
        interval=0.1,
        max_tries=2,
    )
    async def _stream_url_buckets(
        self,
        selection_query: str,
        data_format: str,
        title: Optional[str],
        as_signed_url: Optional[bool],
    ) -> AsyncGenerator[StreamInfoUrl, None]:
        """Get a list of files back for a request

        Args:
            selection_query (str): The selection query we are to do
            data_format (str): The requested file format
            title (Optional[str]): The title of transform to pass to ServiceX
            as_signed_url (Optional[bool]): Return the uri as a presigned http url?

        Yields:
            AsyncIterator[Dict[str, str]]: A tuple of the minio bucket and file in that bucket.
                                           The dict will have entries for:
                                             bucket: The minio bucket name
                                             file: the completed file in the bucket
        """
        query = self._build_json_query(selection_query, data_format, title)

        async with aiohttp.ClientSession() as client:
            # Get a request id - which might be cached, but if not, submit it.
            request_id = await self._get_request_id(client, query)

            # Make sure cache status exists (user could have deleted, see #176)
            if not self._cache.query_status_exists(request_id):
                await self._update_query_status(client, request_id)

            # Get the minio adaptor we are going to use for downloading.
            minio_adaptor = self._minio_adaptor.from_best(
                self._cache.lookup_query_status(request_id)
            )

            # Look up the cache, and then fetch an iterator going through the results
            # from either servicex or the cache, depending.
            try:
                notifier = self._create_notifier(title, False)
                minio_files = self._get_minio_bucket_files_from_servicex(
                    request_id, client, minio_adaptor, notifier
                )

                # Reflect the files back up a level.
                async for object_name in minio_files:
                    uri = (
                        minio_adaptor.get_access_url(request_id, object_name)
                        if as_signed_url
                        else minio_adaptor.get_s3_uri(request_id, object_name)
                    )

                    yield StreamInfoUrl(object_name, uri, request_id)

                # Cache the final status
                await self._update_query_status(client, request_id)

            except ServiceXUnknownRequestID as e:
                self._cache.remove_query(query)
                raise ServiceXUnknownDataRequestID(
                    "Expected the ServiceX backend to know about "
                    f"query {request_id}. It did not. Cleared local"
                    "cache. Please resubmit to trigger a new "
                    "query."
                ) from e

            except ServiceXFatalTransformException as e:
                transform_status = await self._servicex_adaptor.get_query_status(
                    client, request_id
                )
                self._cache.remove_query(query)
                raise ServiceXFatalTransformException(
                    f'ServiceX Fatal Error: {transform_status["failure-info"]}'
                ) from e

            except ServiceXFailedFileTransform as e:
                self._cache.remove_query(query)
                await self._servicex_adaptor.dump_query_errors(client, request_id)
                raise ServiceXException(
                    f"Failed to transform all files in {request_id}"
                ) from e

    @on_exception(backoff.constant, ServiceXUnknownRequestID, interval=0.1, max_tries=3)
    @on_exception(
        backoff.constant,
        (ServiceXUnknownDataRequestID, minio.error.NoSuchBucket),
        interval=0.1,
        max_tries=2,
    )
    async def _data_return(
        self,
        selection_query: str,
        converter: Callable[[Path], Awaitable[Any]],
        title: Optional[str],
        data_format: str = "root-file",
    ) -> List[Any]:
        """Given a query, return the data, in a unique order, that hold
        the data for the query.

        For certian types of exceptions, the queries will be repeated. For example,
        if `ServiceX` indicates that it was restarted in the middle of the query, then
        the query will be re-submitted.

        Arguments:

            selection_query     `qastle` data that makes up the selection request.
            converter           A `Callable` that will convert the data returned from
                                `ServiceX` as a set of files.
            title               Title to send to the backend service
            data_format         The data format that we want to render in

        Returns:

            data                Data converted to the "proper" format, depending
                                on the converter call.
        """
        all_data = {
            f.file: f.data
            async for f in self._stream_return(
                selection_query, title, converter, data_format
            )
        }

        # Finally, we need them in the proper order so we append them
        # all together
        ordered_data = [all_data[k] for k in sorted(all_data.keys())]

        return ordered_data

    async def _stream_return(
        self,
        selection_query: str,
        title: Optional[str],
        converter: Callable[[Path], Awaitable[Any]],
        data_format: str = "root-file",
    ) -> AsyncIterator[StreamInfoData]:
        """Given a query, return the data, in the order it arrives back
        converted as appropriate.

        For certian types of exceptions, the queries will be repeated. For example,
        if `ServiceX` indicates that it was restarted in the middle of the query, then
        the query will be re-submitted.

        Arguments:

            selection_query     `qastle` data that makes up the selection request.
            converter           A `Callable` that will convert the data returned from
                                `ServiceX` as a set of files.

        Returns:

            data                Data converted to the "proper" format, depending
                                on the converter call.
        """
        as_data = (
            StreamInfoData(f.file, await asyncio.ensure_future(converter(f.path)))
            async for f in self._stream_local_files(selection_query, title, data_format)
        )  # type: ignore

        async for r in as_data:
            yield r

    async def _stream_local_files(
        self, selection_query: str, title: Optional[str], data_format: str = "root-file"
    ) -> AsyncGenerator[StreamInfoPath, None]:
        """
        Given a query, return the data as a list of paths pointing to local files
        that contain the results of the query. This is an async generator, and files
        are returned as they arrive.

        For certian types of exceptions, the queries will be repeated. For example,
        if `ServiceX` indicates that it was restarted in the middle of the query, then
        the query will be re-submitted.

        Arguments:

            selection_query     `qastle` data that makes up the selection request.

        Returns:

            data                Data converted to the "proper" format, depending
                                on the converter call.
        """
        # Get a notifier to update anyone who wants to listen.
        notifier = self._create_notifier(title, True)

        # Get all the files
        as_files = (
            f
            async for f in self._get_files(
                selection_query, data_format, notifier, title
            )
        )  # type: ignore

        async for name, a_path in as_files:
            yield StreamInfoPath(name, Path(await a_path))

    async def _get_files(
        self,
        selection_query: str,
        data_format: str,
        notifier: _status_update_wrapper,
        title: Optional[str],
    ) -> AsyncIterator[Tuple[str, Awaitable[Path]]]:
        """
        Return a list of files from servicex as they have been downloaded to this machine. The
        return type is an awaitable that will yield the path to the file.

        For certian types of `ServiceX` failures we will automatically attempt a few retries:

            - When `ServiceX` forgets the query. This sometimes happens when a user submits a
              query, and then disconnects from the network, `ServiceX` is restarted, and then the
              user attempts to download the files from that "no-longer-existing" request.

        Up to 3 re-tries are attempted automatically.

        Arguments:

            selection_query             The query string to send to ServiceX
            data_format                 The type of data that we want to come back.
            notifier                    Status callback to let our progress be advertised
            title                       Title to pass to servicex backend.

        Returns
            Awaitable[Path]             An awaitable that is a path. When it completes, the
                                        path will be valid and point to an existing file.
                                        This is returned this way so a number of downloads can run
                                        simultaneously.
        """
        query = self._build_json_query(selection_query, data_format, title)

        async with aiohttp.ClientSession() as client:
            # Get a request id - which might be cached, but if not, submit it.
            request_id = await self._get_request_id(client, query)

            # Make sure cache status exists (user could have deleted, see #176)
            if not self._cache.query_status_exists(request_id):
                await self._update_query_status(client, request_id)

            # Get the minio adaptor we are going to use for downloading.
            minio_adaptor = self._minio_adaptor.from_best(
                self._cache.lookup_query_status(request_id)
            )

            # Look up the cache, and then fetch an iterator going thorugh the results
            # from either servicex or the cache, depending.
            try:
                cached_files = self._cache.lookup_files(request_id)
                stream_local_files = (
                    self._get_cached_files(cached_files, notifier)
                    if cached_files is not None
                    else self._get_files_from_servicex(
                        request_id, client, minio_adaptor, notifier
                    )
                )

                # Reflect the files back up a level.
                async for r in stream_local_files:
                    yield r

                # Cache the final status
                if cached_files is None:
                    await self._update_query_status(client, request_id)

            except ServiceXUnknownRequestID as e:
                self._cache.remove_query(query)
                raise ServiceXUnknownDataRequestID(
                    "Expected the ServiceX backend to know about "
                    f"query {request_id}. It did not. Cleared local"
                    " cache. Please resubmit to trigger a new "
                    "query."
                ) from e
            except ServiceXNoFilesInCache as e:
                self._cache.remove_query(query)
                raise ServiceXUnknownDataRequestID(
                    "Expected the ServiceX backend to have cached data "
                    f"from query {request_id}. It did not. Cleared local"
                    " cache. Please resubmit to trigger a new "
                    "query."
                ) from e

            except ServiceXFatalTransformException as e:
                transform_status = await self._servicex_adaptor.get_query_status(
                    client, request_id
                )
                self._cache.remove_query(query)
                raise ServiceXFatalTransformException(
                    f'ServiceX Fatal Error: {transform_status["failure-info"]}'
                ) from e

            except ServiceXFailedFileTransform as e:
                self._cache.remove_query(query)
                await self._servicex_adaptor.dump_query_errors(client, request_id)
                raise ServiceXException(
                    f"Failed to transform all files in {request_id}"
                ) from e

    async def _get_request_id(
        self, client: aiohttp.ClientSession, query: Dict[str, Any]
    ) -> str:
        """
        For this query, fetch the request id. If we have it cached, use that. Otherwise, query
        ServiceX for a enw one (and cache it for later use).
        """
        request_id = self._cache.lookup_query(query)
        if request_id is None:
            request_info = await self._servicex_adaptor.submit_query(client, query)
            request_id = request_info["request_id"]
            self._cache.set_query(query, request_id)
            await self._update_query_status(client, request_id)
        return request_id

    async def _update_query_status(
        self, client: aiohttp.ClientSession, request_id: str
    ):
        """Fetch the status from servicex and cache it locally, over
        writing what was there before.

        Args:
            request_id (str): Request id of the status to fetch and cache.
        """
        info = await self._servicex_adaptor.get_query_status(client, request_id)
        self._cache.set_query_status(info)

    async def _get_cached_files(
        self, cached_files: List[Tuple[str, Path]], notifier: _status_update_wrapper
    ):
        """
        Return the list of files as an iterator that we have pulled from the cache
        """
        notifier.update(processed=len(cached_files), remaining=0, failed=0)
        loop = asyncio.get_event_loop()
        for f, p in cached_files:
            path_future = loop.create_future()
            path_future.set_result(p)
            notifier.inc(downloaded=1)
            yield f, path_future

    async def _download_a_file(
        self,
        stream: AsyncIterator[str],
        request_id: str,
        minio_adaptor: MinioAdaptor,
        notifier: _status_update_wrapper,
    ) -> AsyncIterator[Tuple[str, Awaitable[Path]]]:
        """
        Given an object name and request id, fetch the data locally from the minio bucket.
        The copy can take a while, so send it off to another thread - don't pause queuing up other
        files to download.
        """

        async def do_copy(final_path):
            assert request_id is not None
            await minio_adaptor.download_file(request_id, f, final_path)
            notifier.inc(downloaded=1)
            notifier.broadcast()
            return final_path

        file_object_list: List[Tuple[str, Path]] = []
        async for f in stream:
            copy_to_path = self._cache.data_file_location(request_id, f)
            file_object_list.append((f, copy_to_path))

            yield f, do_copy(copy_to_path)

        self._cache.set_files(request_id, file_object_list)

    async def _get_files_from_servicex(
        self,
        request_id: str,
        client: aiohttp.ClientSession,
        minio_adaptor: MinioAdaptor,
        notifier: _status_update_wrapper,
    ):
        """
        Fetch query result files from `servicex`. Given the `request_id` we will download
        files as they become available. We also coordinate caching.
        """
        start_time = time.monotonic()
        good = True
        try:
            # Get the stream of minio bucket new files.
            stream_new_object = self._get_minio_bucket_files_from_servicex(
                request_id, client, minio_adaptor, notifier
            )

            # Next, download the files as they are found (and return them):
            stream_downloaded = self._download_a_file(
                stream_new_object, request_id, minio_adaptor, notifier
            )

            # Return the files to anyone that wants them!

            async for info in stream_downloaded:
                yield info

        except Exception:
            good = False
            raise

        finally:
            end_time = time.monotonic()
            run_time = timedelta(seconds=end_time - start_time)
            logging.getLogger(__name__).info(
                f"Running servicex query for " f"{request_id} took {run_time}"
            )
            self._log.write_query_log(
                request_id,
                notifier.total,
                notifier.failed,
                run_time,
                good,
                self._cache.path,
            )

    async def _get_minio_bucket_files_from_servicex(
        self,
        request_id: str,
        client: aiohttp.ClientSession,
        minio_adaptor: MinioAdaptor,
        notifier: _status_update_wrapper,
    ) -> AsyncIterator[str]:
        """Create an async stream of `minio` bucket/filenames from a request id.

        Args:
            request_id (str): The request id that we should be polling for updates.
            client (aiohttp.ClientSession): The client connection to make API queries on
            minio_adaptor (MinioAdaptor): The minio adaptor we can use to connect to the minio
                                          bucket for new items.
            notifier (_status_update_wrapper): Allows us to send updates of progress
                                               back to the user

        Yields:
            [type]: Returns xxx and yyy.
        """
        start_time = time.monotonic()
        try:
            # Setup the status sequence from servicex
            stream_status = transform_status_stream(
                self._servicex_adaptor, client, request_id
            )
            stream_notified = stream_status_updates(stream_status, notifier)
            stream_watched = trap_servicex_failures(stream_notified)
            stream_unique = stream_unique_updates_only(stream_watched)

            # Next, download the files as they are found (and return them):
            stream_new_object = find_new_bucket_files(
                minio_adaptor, request_id, stream_unique
            )

            # Return the minio information.
            async for info in stream_new_object:
                yield info

        finally:
            end_time = time.monotonic()
            run_time = timedelta(seconds=end_time - start_time)
            logging.getLogger(__name__).info(
                f"Running servicex query for "
                f"{request_id} took {run_time} (no files downloaded)"
            )

    def _build_json_query(
        self, selection_query: str, data_format: str, title: Optional[str]
    ) -> Dict[str, Union[str, Iterable[str]]]:
        """
        Returns a list of locally written files for a given selection query.

        Arguments:
            selection_query         The query to be send into the ServiceX API
            data_format             What is the output data type (parquet, root-file, etc.)

        Notes:
            - Internal routine.
        """
        assert data_format in g_allowed_formats

        # Items that must always be present
        json_query: Dict[str, Union[str, Iterable[str]]] = {
            "selection": selection_query,
            "result-destination": self._result_destination,
            "result-format": "parquet" if data_format == "parquet" else "root-file",
            "chunk-size": "1000",
            "workers": str(self._max_workers),
        }

        # Add the appropriate did.
        # Capture full did as well as single item files (see  #178)
        if isinstance(self._dataset, str):
            if self._dataset[0:7].lower() in ["root://", "http://"]:
                json_query["file-list"] = [self._dataset]
            else:
                json_query["did"] = self._dataset
        else:
            json_query["file-list"] = self._dataset

        # Optional items
        if self._codegen is not None:
            json_query["codegen"] = self._codegen

        if self._image is not None:
            json_query["image"] = self._image

        if title is not None:
            json_query["title"] = title

        logging.getLogger(__name__).debug(
            f"JSON to be sent to servicex: {str(json_query)}"
        )

        return json_query

    # Define the synchronous versions of the async methods for easy of use

    get_data_rootfiles_uri = make_sync(get_data_rootfiles_uri_async)
    get_data_parquet_uri = make_sync(get_data_parquet_uri_async)



import hashlib
import json
import logging
import os
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Dict, List, Optional, Set, Tuple

from .utils import ServiceXException, _query_cache_hash, sanitize_filename

_ignore_cache = False


@contextmanager
def ignore_cache():
    """This will cause all caches to be ignored while it is invoked:

    ```
    with ignore_cache():
        ServiceXDataset(...).get_data...()
    ```

    If you want to do this globally, you can just use the `__enter__()` method.
    This is probably the only way to do this accross cells in a notebook.

    ```
    i = ignore_cache()
    i.__enter__()
    ... Query code, jupyter notebook cells, etc. go here
    i.__exit(None, None, None)
    ```

    Note:

    - The only time the cache is checked is when the query is actually made, not when
      the servicex dataset object is created!
    - Calls to this can be safely nested.
    - Note that calling this doesn't clear the cache or delete anything. It
      just prevents the cache lookup from working while it is in effect.
    """
    global _ignore_cache
    old_value = _ignore_cache
    _ignore_cache = True
    yield
    _ignore_cache = old_value


# Track where we should write out an analysis cache (and search up-stream).
# This contains the directory of the file.
_g_analysis_cache_location: Optional[Path] = None
_g_analysis_cache_filename: str = "servicex_query_cache.json"

# List of queries we know to be bad from this run
_g_bad_query_cache_ids: Set[str] = set()


def reset_local_query_cache():
    """Used to reset the analysis cache location. Normally called only
    during testing.
    """
    global _g_analysis_cache_location
    _g_analysis_cache_location = None
    global _g_analysis_cache_filename
    _g_analysis_cache_filename = "servicex_query_cache.json"
    global _g_bad_query_cache_ids
    _g_bad_query_cache_ids = set()


reset_local_query_cache()


def update_local_query_cache(analysis_cache: Optional[Path] = None):
    """Record analysis query hashes in an analysis cache.

    If this routine is not called, the current directory is searched for an
    analysis cache file. If found, it participates in the query lookup.

    After this routine is called, then when a query is made or looked up, an
    analysis cache file in the local directory is updated with new query request
    id's.

    This will allow one user to send a file to another user, along with the
    servicex backend in `servicex.yaml` allow them to fetch the same data.
    (or share on a similar machine).

    Args:
        analysis_cache (Optional[Path], optional): The directory or filename of
        the analysis cache file. If `None` defaults to the file `servicex_query_cache.json` in
        the local directory. If only a directory is passed, then the `servicex_query_cache.json`
        in that directory is used. Defaults to None.
    """
    file_path = (
        Path(".")
        if analysis_cache is None
        else analysis_cache
        if analysis_cache.is_dir()
        else analysis_cache.parent
    )

    name = (
        "servicex_query_cache.json"
        if (analysis_cache is None or analysis_cache.is_dir())
        else analysis_cache.name
    )

    global _g_analysis_cache_filename
    global _g_analysis_cache_location
    if _g_analysis_cache_location is not None and _g_analysis_cache_filename != name:
        raise ServiceXException(
            "Updating local query cache called twice, with "
            f"{_g_analysis_cache_filename} and {name}."
        )
    _g_analysis_cache_filename = name

    if (
        _g_analysis_cache_location is not None
        and _g_analysis_cache_location != file_path
    ):
        raise ServiceXException(
            "Updating local query cache called twice, with "
            f"{_g_analysis_cache_location} and {file_path}."
        )
    _g_analysis_cache_location = file_path

    loc = _g_analysis_cache_location / _g_analysis_cache_filename
    if not loc.exists():
        loc.parent.mkdir(parents=True, exist_ok=True)
        loc.touch()


class Cache:
    """
    Caching for all data returns from the system. It provides both in-memory
    and on-disk cache.

    TODO: Rename this to be an adaptor, unifying how we name things
    """

    _in_memory_cache = {}

    @classmethod
    def reset_cache(cls):  # # pragma: no cover
        "Reset the internal cache, usually used for testing"
        cls._in_memory_cache = {}

    def __init__(
        self,
        cache_path: Path,
        ignore_cache: bool = False,
        analysis_query_key: str = "default",
    ):
        """
        Create the cache object

        Arguments:

            cache_path          The path to the cache directory. Only sub-directories
                                will be created in this path.
            ignore_cache        If true, then always ignore the cache for any queries
                                against this dataset.
        """
        self._path = cache_path
        self._ignore_cache = ignore_cache
        self._analysis_query_key = analysis_query_key

        # old windows versions have a limit of file path of 260 chars.
        self.max_path_len = 200
        if os.name == "nt":
            self.max_path_len = self.max_path_len - 100

    @property
    def path(self) -> Path:
        "Return root path of cache directory"
        return self._path

    @contextmanager
    def ignore_cache(self):
        """Ignore the cache as long as we are held. Supports nesting."""
        old_ignore = self._ignore_cache
        self._ignore_cache = True
        yield
        self._ignore_cache = old_ignore

    def _query_cache_file(self, json: Dict[str, str]) -> Path:
        "Return the query cache file"
        h = _query_cache_hash(json)
        return self._path / "query_cache" / h

    def _query_status_cache_file(self, request_id: str) -> Path:
        "Return the query cache file"
        return self._path / "query_cache_status" / request_id

    def _files_cache_file(self, id: str) -> Path:
        "Return the file that contains the list of files"
        return self._path / "file_list_cache" / id

    def lookup_query(self, json: Dict[str, str]) -> Optional[str]:
        global _ignore_cache
        if _ignore_cache or self._ignore_cache:
            return None

        f = self._query_cache_file(json)
        if not f.exists():
            hash = _query_cache_hash(json)
            if hash in _g_bad_query_cache_ids:
                return None
            return self._lookup_analysis_query_cache(hash)

        with f.open("r") as i:
            request_id = i.readline().strip()
            logging.getLogger(__name__).debug(f"Found query '{request_id}' in cache")

        self._write_analysis_query_cache(json, request_id)

        return request_id

    def set_query(self, json: Dict[str, str], v: str):
        """Associate a query with a request-id.

        A hash is taken of the query.

        Args:
            json (Dict[str, str]): The query JSON
            v (str): The `request-id`
        """
        f = self._query_cache_file(json)
        f.parent.mkdir(parents=True, exist_ok=True)
        with f.open("w") as o:
            o.write(f"{v}\n")

        self._write_analysis_query_cache(json, v)

    def remove_query(self, json: Dict[str, Any]):
        """Remove the query from our local and analysis caches

        Args:
            json (Dict[str, Any]): The query to remove
        """
        f = self._query_cache_file(json)
        if f.exists():
            f.unlink()
        self._remove_from_analysis_cache(_query_cache_hash(json))

    def _load_full_analysis_query_cache(self) -> Optional[Dict[str, Dict[str, str]]]:
        """Safely load the analysis query cache, with all elements of the cache

        - If there is no cache file, return None
        - If the file is empty, return an empty cache
        - Return the contents of the file.

        Returns:
            Optional[Dict[str, str]]: Returns the query cache if it exists
        """
        q_file = self._get_analysis_cache_file()
        if not q_file.exists():
            return None

        analysis_cache = {}
        with q_file.open("r") as input:
            text = input.read()
            if len(text) > 0:
                analysis_cache: Dict[str, Dict[str, str]] = json.loads(text)

        return analysis_cache

    def _load_analysis_query_cache(self) -> Optional[Dict[str, str]]:
        """Safely load the analysis query cache.

        - If there is no cache file, return None
        - If the file is empty, return an empty cache
        - Return the contents of the file.

        Returns:
            Optional[Dict[str, str]]: Returns the query cache if it exists
        """
        full_analysis_cache = self._load_full_analysis_query_cache()
        if full_analysis_cache is None:
            return None

        # Find "our" analysis cache
        analysis_cache = {}
        if self._analysis_query_key in full_analysis_cache:
            analysis_cache = full_analysis_cache[self._analysis_query_key]

        return analysis_cache

    def _save_analysis_query_cache(self, cache: Dict[str, str]):
        full_cache = self._load_full_analysis_query_cache()
        assert (
            full_cache is not None
        ), "Internal error, should have been checked already"

        full_cache[self._analysis_query_key] = cache

        cache_file = self._get_analysis_cache_file()
        with cache_file.open("w") as output:
            json.dump(full_cache, output)

    def _write_analysis_query_cache(self, query_info: Dict[str, str], request_id: str):
        """Write out a local analysis query hash-request-id assocaition.

        Args:
            query_info (Dict[str, str]): The JSON of the request
            request_id (str): The `request-id`
        """
        analysis_cache = self._load_analysis_query_cache()
        if analysis_cache is None:
            return

        analysis_cache[_query_cache_hash(query_info)] = request_id

        self._save_analysis_query_cache(analysis_cache)

    def _remove_from_analysis_cache(self, query_hash: str):
        """Remove an item from the analysis cache if we are writing to it!

        Args:
            query_hash (str): The hash we will remove
        """
        if _g_analysis_cache_location is None:
            _g_bad_query_cache_ids.add(query_hash)
            return

        cache_contents = self._find_analysis_cached_query(query_hash)
        if cache_contents is not None:
            del cache_contents[query_hash]
            self._save_analysis_query_cache(cache_contents)

    def _lookup_analysis_query_cache(
        self,
        query_hash: str,
        filename: Optional[str] = None,
        location: Optional[Path] = None,
    ) -> Optional[str]:
        """Look at all possible query caches for this query.

        If `location` is `None`, then start from the global location searching for a query file.
        If `location` is specified, check that directory.
        In both cases, if the query hash isn't found, then move up one directory and try again.

        `filename` is the name of the file we should be looking for. If `None` default to the
        global.

        Args:
            query_hash (str): The hash of the query we need to lookup.
            filename (Optional[str]): The name fo the file that contains the cache. If not
            specified then defaults to the global.
            location (Optional[Path]): Directory to start searching in. If not specified then
            defaults to the global. If that isn't specified, defaults to the current directory.

        Returns:
            (Optional[str]): The return hash of what we need to look up
        """
        cache_contents = self._find_analysis_cached_query(query_hash)
        if cache_contents is not None:
            return cache_contents[query_hash]
        return None

    def _get_analysis_cache_file(
        self, filename: Optional[str] = None, location: Optional[Path] = None
    ) -> Path:
        """Get our best guess as to where the analysis cache file will be

        It will use the globally set defaults if nothing is specified.

        Args:
            filename (Optional[str], optional): Cache filename to use. Defaults to None.
            location (Optional[str], optional): Cache location to use. Defaults to None.
        """
        c_filename = filename if filename is not None else _g_analysis_cache_filename
        c_location = (
            location
            if location is not None
            else _g_analysis_cache_location
            if _g_analysis_cache_location is not None
            else Path(".")
        )

        # Now see if we can find it
        loc = c_location.absolute() / c_filename
        while (not loc.exists()) and (len(loc.parts) >= 3):
            new_parent = loc.parent.parent
            loc = new_parent / c_filename

        return loc

    def _find_analysis_cached_query(
        self,
        query_hash: str,
        filename: Optional[str] = None,
        location: Optional[Path] = None,
    ) -> Optional[Dict[str, str]]:
        """Returns the contents of an analysis cache file and the file that contains
        a query hash

        Args:
            has (str): The hash of the query we are to find

        Returns:
            Tuple[Dict[str, str], Path]: The contents of the file and the path to the
            analysis cache that contains the hash. `None` if the query was not found
        """
        # If the cache is here, then see if it has a hit
        analysis_cache = self._load_analysis_query_cache()
        if analysis_cache is not None:
            if query_hash in analysis_cache:
                return analysis_cache

        return None

    def set_query_status(self, query_info: Dict[str, str]):
        """Cache a query status (json dict)

        Args:
            query_info (Dict[str, str]): The info we should cache. Must contain `request_id`.
        """
        assert (
            "request_id" in query_info
        ), "Internal error - request_id should always be part of info returned"
        f = self._query_status_cache_file(query_info["request_id"])
        f.parent.mkdir(parents=True, exist_ok=True)
        with f.open("w") as o:
            json.dump(query_info, o)

    def lookup_query_status(self, request_id: str) -> Dict[str, str]:
        """Returns the info from the last time the query status was cached.

        Args:
            request_id (str): Request id we should look up.

        """
        f = self._query_status_cache_file(request_id)
        if not f.exists():
            raise ServiceXException(f"No cache information for query {request_id}")
        with f.open("r") as o:
            return json.load(o)

    def query_status_exists(self, request_id: str) -> bool:
        """Returns true if the query status file exists on the local machine.

        Args:
            request_id (str): The request-id to look up

        Returns:
            bool: True if present, false otherwise.
        """
        return self._query_status_cache_file(request_id).exists()

    def set_files(self, id: str, files: List[Tuple[str, Path]]):
        """Cache the files for this request

        Note: We do check to make sure all the files exist

        Args:
            id (str): The request-id
            files (List[Tuple[str, Path]]): the minio buck name and local file paths
        """
        # Make sure they exist
        assert all(f.exists() for _, f in files)
        f = self._files_cache_file(id)
        f.parent.mkdir(parents=True, exist_ok=True)
        with f.open("w") as o:
            json.dump([(n, str(f)) for n, f in files], o)

    def lookup_files(self, id: str) -> Optional[List[Tuple[str, Path]]]:
        """Return a list of files in the cache for a request id.

        - Returns None if there is nothing in the cache
        - Returns None if any of the files are missing

        Args:
            id (str): Request-id we are looking up

        Returns:
            Optional[List[Tuple[str, Path]]]: List of minio-bucket to local file mappings
        """
        f = self._files_cache_file(id)
        if not f.exists():
            return None
        with f.open("r") as i:
            list_of_cached_files = [(n, Path(p)) for n, p in json.load(i)]
        if all(f.exists() for _, f in list_of_cached_files):
            return list_of_cached_files
        return None

    def set_inmem(self, id: str, v: Any):
        self._in_memory_cache[id] = v

    def lookup_inmem(self, id: str) -> Optional[Any]:
        global _ignore_cache
        if _ignore_cache or self._ignore_cache:
            return None

        if id not in self._in_memory_cache:
            return None
        return self._in_memory_cache[id]

    def data_file_location(self, request_id: str, data_name: str) -> Path:
        """
        Return the path to the file that should be written out for this
        data_name. This is where the output file should get stored.
        Truncate the leftmost characters from filenames to avoid throwing a
        OSError: [Errno 63] File name too long error. Use a hash string to
        make sure that the file names remain unique.
        """
        parent = self._path / "data" / request_id
        parent.mkdir(parents=True, exist_ok=True)
        sanitized = sanitize_filename(data_name)
        if (len(sanitized) + len(parent.name)) > self.max_path_len:
            hash = hashlib.md5(sanitized.encode())
            hash_string = hash.hexdigest()
            max_len = self.max_path_len - len(parent.name) - len(hash_string) - 1
            sanitized = f"{hash_string}-{sanitized[len(sanitized)-max_len:]}"

        return parent / sanitized



from pathlib import Path

import minio
from minio.error import ResponseError
import pytest
from urllib3.exceptions import MaxRetryError
from minio.error import NoSuchBucket

from servicex import (
    MinioAdaptor,
    ServiceXException,
)
from servicex.minio_adaptor import MinioAdaptorFactory, find_new_bucket_files
from servicex.utils import ServiceXNoFilesInCache


def make_minio_file(fname):
    from unittest import mock

    r = mock.MagicMock()
    r.object_name = fname
    return r


@pytest.fixture
def good_minio_client(mocker):

    minio_client = mocker.MagicMock(spec=minio.Minio)
    minio_client.list_objects.return_value = [
        make_minio_file(
            "root:::dcache-atlas-xrootd-wan.desy.de:1094::pnfs:desy.de:atlas"
            ":dq2:atlaslocalgroupdisk:rucio:mc15_13TeV:8a:f1:DAOD_STDM3.05630052"
            "._000001.pool.root.198fbd841d0a28cb0d9dfa6340c890273-1.part.minio"
        )
    ]
    minio_client.presigned_get_object.return_value = "http://data.done"

    mocker.patch("servicex.minio_adaptor.Minio", return_value=minio_client)

    p_rename = mocker.patch("servicex.minio_adaptor.Path.rename", mocker.MagicMock())

    mocker.patch("servicex.minio_adaptor.Path.mkdir", mocker.MagicMock())

    return p_rename, minio_client


@pytest.fixture
def bad_minio_client(mocker):
    minio_client = mocker.MagicMock(spec=minio.Minio)
    minio_client.list_objects.return_value = [
        make_minio_file(
            "root:::dcache-atlas-xrootd-wan.desy.de:1094::pnfs:desy.de:atlas:dq2"
            ":atlaslocalgroupdisk:rucio:mc15_13TeV:8a:f1:DAOD_STDM3.05630052._000001"
            ".pool.root.198fbd841d0a28cb0d9dfa6340c890273-1.part.minio"
        )
    ]
    minio_client.fget_object.side_effect = Exception("this copy really failed")

    mocker.patch("servicex.minio_adaptor.Minio", return_value=minio_client)

    p_rename = mocker.patch("servicex.minio_adaptor.Path.rename", mocker.MagicMock())
    mocker.patch("servicex.minio_adaptor.Path.mkdir", mocker.MagicMock())

    return p_rename, minio_client


@pytest.fixture
def bad_minio_endpoint(mocker):
    "Simulate the Minio end point not being there at all"
    response1 = mocker.MagicMock()
    response1.data = "<xml></xml>"

    minio_client = mocker.MagicMock(spec=minio.Minio)
    pool = mocker.MagicMock()
    pool.host = "http://dude"
    pool.port = "fork-me"
    minio_client.list_objects.side_effect = MaxRetryError(pool, "http://dude")

    mocker.patch("servicex.minio_adaptor.Minio", return_value=minio_client)

    p_rename = mocker.patch("servicex.minio_adaptor.Path.rename", mocker.MagicMock())
    mocker.patch("servicex.minio_adaptor.Path.mkdir", mocker.MagicMock())

    return p_rename, minio_client


@pytest.fixture
def minio_endpoint_bucket_missing(mocker):
    "Simulate the Minio end point throwing a no files in bucket error"
    response1 = mocker.MagicMock()
    response1.data = "<xml></xml>"

    minio_client = mocker.MagicMock(spec=minio.Minio)
    pool = mocker.MagicMock()
    pool.host = "http://dude"
    pool.port = "fork-me"
    minio_client.list_objects.side_effect = NoSuchBucket(200)

    mocker.patch("servicex.minio_adaptor.Minio", return_value=minio_client)

    p_rename = mocker.patch("servicex.minio_adaptor.Path.rename", mocker.MagicMock())
    mocker.patch("servicex.minio_adaptor.Path.mkdir", mocker.MagicMock())

    return p_rename, minio_client


@pytest.fixture
def bad_then_good_minio_listing(mocker):
    "Simulate the Minio going offline and then coming back"
    response1 = mocker.MagicMock()
    response1.data = "<xml></xml>"
    response2 = [
        make_minio_file(
            "root:::dcache-atlas-xrootd-wan.desy.de:1094::pnfs:desy.de:atlas"
            ":dq2:atlaslocalgroupdisk:rucio:mc15_13TeV:8a:f1:DAOD_STDM3."
            "05630052._000001.pool.root.198fbd841d0a28cb0d9dfa6340c890273-1"
            ".part.minio"
        )
    ]

    minio_client = mocker.MagicMock(spec=minio.Minio)
    minio_client.list_objects.side_effect = [
        ResponseError(response1, "POST", "Due"),
        response2,
    ]

    mocker.patch("servicex.minio_adaptor.Minio", return_value=minio_client)

    p_rename = mocker.patch("servicex.minio_adaptor.Path.rename", mocker.MagicMock())
    mocker.patch("servicex.minio_adaptor.Path.mkdir", mocker.MagicMock())

    return p_rename, minio_client


@pytest.mark.asyncio
async def test_download_good(good_minio_client):
    ma = MinioAdaptor("localhost:9000")

    final_path = Path("/tmp/output-file.dude")  # type: Path
    await ma.download_file("111-22-333-444", "dude-where-is-my-lunch", final_path)

    p_rename, p_minio = good_minio_client

    # Make sure copy was called.
    p_minio.fget_object.assert_called_once()
    expected_path = Path("/tmp/output-file.dude.temp")
    p_minio.fget_object.assert_called_with(
        "111-22-333-444", "dude-where-is-my-lunch", str(expected_path)
    )

    # Make sure the rename was done.
    p_rename.assert_called_with(final_path)


@pytest.mark.asyncio
async def test_download_bad(bad_minio_client):
    ma = MinioAdaptor("localhost:9000")

    final_path = Path("/tmp/output-file.dude")
    with pytest.raises(ServiceXException) as e:
        await ma.download_file("111-22-333-444", "dude-where-is-my-lunch", final_path)

    assert "Failed to copy" in str(e.value)

    p_rename, p_minio = bad_minio_client
    p_minio.fget_object.assert_called_once()

    # Make sure the rename was done.
    p_rename.assert_not_called()


@pytest.mark.asyncio
async def test_download_already_there(mocker, good_minio_client):
    p_exists = mocker.patch("servicex.minio_adaptor.Path.exists", return_value=True)

    ma = MinioAdaptor("localhost:9000")

    final_path = Path("/tmp/output-file.dude")  # type: Path
    await ma.download_file("111-22-333-444", "dude-where-is-my-lunch", final_path)

    p_rename, p_minio = good_minio_client

    # Make sure copy was called.
    p_minio.fget_object.assert_not_called()

    # Make sure the rename was done.
    p_rename.assert_not_called()

    # Make sure the exists worked.
    p_exists.assert_called_once()


def test_access_url(good_minio_client):
    "Make sure the presigned_get_object is properly called"
    mn = MinioAdaptor("localhost:9000")
    assert mn.get_access_url("123-456", "file1") == "http://data.done"
    good_minio_client[1].presigned_get_object.assert_called_with("123-456", "file1")


def test_access_s3_uri():
    mn = MinioAdaptor("localhost:9000")
    assert mn.get_s3_uri("123-456", "file1") == "s3://123-456/file1"


def test_list_objects(good_minio_client):
    ma = MinioAdaptor("localhost:9000")

    f = ma.get_files("111-222-333-444")
    assert len(f) == 1  # type: ignore


def test_list_objects_with_null(bad_then_good_minio_listing):
    """Sometimes for reasons we do not understand we get back a response error from
    list_objects minio method"""
    ma = MinioAdaptor("localhost:9000")
    f = ma.get_files("111-222-333-444")
    assert len(f) == 1  # type: ignore


def test_bad_minio(bad_minio_endpoint):
    """Sometimes for reasons we do not understand we get back a response error from
    list_objects minio method"""
    ma = MinioAdaptor("localhost:9000")
    with pytest.raises(ServiceXException):
        ma.get_files("111-222-333-444")


def test_no_files_found(minio_endpoint_bucket_missing):
    """When minio doesn't have any files in a bucket, we
    throw an error - make sure it gets translated into an SX
    error that can cause a re-try.
    """
    ma = MinioAdaptor("localhost:9000")
    with pytest.raises(ServiceXNoFilesInCache):
        ma.get_files("111-222-333-444")


@pytest.mark.asyncio
async def test_find_new_bucket_0_files(mocker):
    from .conftest import MockMinioAdaptor, as_async_seq

    ma = MockMinioAdaptor(mocker)
    r = [
        f
        async for f in find_new_bucket_files(
            ma, "123-456", as_async_seq([(1, 0, 0), (0, 1, 0)])  # type: ignore
        )
    ]

    assert len(r) == 0


@pytest.mark.asyncio
async def test_find_new_bucket_1_files(mocker):
    from .conftest import MockMinioAdaptor, as_async_seq

    ma = MockMinioAdaptor(mocker, files=["one_two_three"])
    r = [
        f
        async for f in find_new_bucket_files(
            ma, "123-456", as_async_seq([(1, 0, 0), (0, 1, 0)])  # type: ignore
        )
    ]

    assert len(r) == 1


def test_factory_no_inputs():
    f = MinioAdaptorFactory()
    with pytest.raises(ServiceXException):
        # Should fail b.c. no way to figure out what to create!
        f.from_best()


def test_factor_always():
    a = MinioAdaptor("localhost:9000")
    f = MinioAdaptorFactory(always_return=a)
    assert f.from_best() is a


def test_factory_from_request():
    info = {
        "minio-access-key": "miniouser",
        "minio-endpoint": "minio.servicex.com:9000",
        "minio-secured": False,
        "minio-secret-key": "leftfoot1",
    }
    m = MinioAdaptorFactory().from_best(info)
    assert m._endpoint == "minio.servicex.com:9000"
    assert not m._secured
    assert m._access_key == "miniouser"
    assert m._secretkey == "leftfoot1"



import asyncio
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from typing import Iterable, Optional

import pandas as pd
import awkward as ak

from .utils import ServiceXException

_conversion_pool = ThreadPoolExecutor(4)


class DataConverterAdaptor:
    """Methods to convert from one type of data to the other."""

    def __init__(self, default_file_type: str):
        """Create a data converter adaptor. By default it will do the
        conversation as requested.

        Args:
            default_file_type (str): The default file type (`parquet` or `root`)
        """
        self._default_file_type = default_file_type

    async def convert_to_pandas(self, file: Path, file_type: Optional[str] = None):
        """Convert to a pandas dataframe from data stored in a file of a particular file_type

        Args:
            file (Path): Path to the file
            file_type (str): What the file contains (root, parquet, etc)
        """
        file_type = file_type if file_type is not None else self._default_file_type
        if file_type == "root":
            return await self._convert_root_to_pandas(file)
        elif file_type == "parquet":
            return await self._convert_parquet_to_pandas(file)
        else:
            raise ServiceXException(
                f"Conversion from {file_type} into an pandas DF is not " "yet supported"
            )

    async def convert_to_awkward(self, file: Path, file_type: Optional[str] = None):
        """Convert to an awkward data array from data stored in a file of a particular file_type

        Args:
            file (Path): Path to the file
            file_type (str): What the file contains (root, parquet, etc)
        """
        file_type = file_type if file_type is not None else self._default_file_type
        if file_type == "root":
            return await self._convert_root_to_awkward(file)
        elif file_type == "parquet":
            return await self._convert_parquet_to_awkward(file)
        else:
            raise ServiceXException(
                f"Conversion from {file_type} into an awkward array is not "
                "yet supported"
            )

    def combine_pandas(self, dfs: Iterable[pd.DataFrame]) -> pd.DataFrame:
        """Combine many pandas dataframes into a single one, in order.

        Args:
            dfs (Iterable[pd.DataFrame]): The list of dataframes
        """
        return pd.concat(dfs)

    def combine_awkward(self, awks: Iterable[ak.Array]) -> ak.Array:
        """Combine many awkward arrays into a single one, in order.

        Args:
            awks (Iterable[ChunkedArray]): The input list of awkward arrays
        """
        return ak.concatenate(awks)  # type: ignore

    async def _convert_root_to_pandas(self, file: Path):
        """
        Convert the contents of a ROOT file to pandas.

        Arguments:

            file        A `Path` to the file containing the pandas data

        Returns:

            DataFrame   A pandas dataframe

        Note:

            - Work is done on a second thread.
            - Pandas is only imported if this is called.

        """
        from pandas import DataFrame

        def do_the_work(file: Path) -> DataFrame:
            import uproot as uproot

            with uproot.open(file) as f_in:
                r = f_in[f_in.keys()[0]]
                return r.arrays(library="pd")  # type: ignore

        return await asyncio.wrap_future(_conversion_pool.submit(do_the_work, file))

    async def _convert_parquet_to_pandas(self, file: Path):
        """
        Convert the contents of a parquet file to pandas.

        Arguments:

            file        A `Path` to the file containing the pandas data

        Returns:

            DataFrame   A pandas dataframe

        Note:

            - Work is done on a second thread.
            - Pandas is only imported if this is called.

        """
        import pandas as pd

        def do_the_work(file: Path) -> pd.DataFrame:
            return pd.read_parquet(str(file))

        return await asyncio.wrap_future(_conversion_pool.submit(do_the_work, file))

    async def _convert_root_to_awkward(self, file: Path):
        """
        Convert the contents of a ROOT file to an awkward dictionary.

        Arguments:

            file        A `Path` to the file containing the pandas data

        Returns:

            DataFrame   A pandas dataframe

        Note:

            - Work is done on a second thread.
            - Awkward is only imported if this is called.
            - A LazyArray is returned, so it isn't completely loaded into memory. That also means
              this will leak filehandles - as that has to be left open.

        """

        def do_the_work(file: Path) -> ak.Array:
            import uproot as uproot

            with uproot.open(file) as f_in:
                tree_name = f_in.keys()[0]

            return uproot.lazy(f"{file}:{tree_name}")

        return await asyncio.wrap_future(_conversion_pool.submit(do_the_work, file))

    async def _convert_parquet_to_awkward(self, file: Path):
        """
        Convert the contents of a parquet file to an awkward dictionary.

        Arguments:

            file        A `Path` to the file containing the pandas data

        Returns:

            DataFrame   A pandas dataframe

        Note:

            - Work is done on a second thread.
            - Pandas is only imported if this is called.

        """

        def do_the_work(file: Path) -> ak.Array:
            # TODO: When we move to awkward1, make sure this becomes lazy
            return ak.from_parquet(str(file))  # type: ignore

        return await asyncio.wrap_future(_conversion_pool.submit(do_the_work, file))



import os
import tempfile
from datetime import timedelta
from pathlib import Path
from typing import Optional

import aiohttp
import backoff
import pytest
from servicex.utils import (
    _query_cache_hash,
    _status_update_wrapper,
    clean_linq,
    dataset_as_name,
    default_client_session,
    get_configured_cache_path,
    log_adaptor,
    on_exception_itr,
    stream_status_updates,
    stream_unique_updates_only,
    write_query_log,
)

from .conftest import as_async_seq


@pytest.fixture
def log_location_file(tmp_path):
    l_file = tmp_path / "log.csv"
    if l_file.exists():
        l_file.unlink()
    yield l_file
    if l_file.exists():
        l_file.unlink()


@pytest.mark.asyncio
async def test_client_session():
    c = await default_client_session()
    assert isinstance(c, aiohttp.ClientSession)


@pytest.mark.asyncio
async def test_client_session_same():
    # On same thread they should be the same.
    c1 = await default_client_session()
    c2 = await default_client_session()

    assert c1 is c2


@pytest.mark.asyncio
async def test_client_session_different_threads():
    async def get_a_client_async():
        return await default_client_session()

    from make_it_sync import make_sync

    get_a_client = make_sync(get_a_client_async)

    c1 = get_a_client()
    c2 = await get_a_client_async()

    assert c1 is not c2


def test_log_file_created(log_location_file):
    write_query_log(
        "123",
        10,
        0,
        timedelta(seconds=20),
        True,
        path_to_log_dir=log_location_file.parent,
    )
    assert log_location_file.exists()
    t = log_location_file.read_text().split("\n")[1]
    assert t == "123,10,0,20.0,1"


def test_log_file_none_files(log_location_file):
    write_query_log(
        "123",
        None,
        0,
        timedelta(seconds=20),
        True,
        path_to_log_dir=log_location_file.parent,
    )
    assert log_location_file.exists()
    t = log_location_file.read_text().split("\n")[1]
    assert t == "123,-1,0,20.0,1"


def test_log_file_write_minutes(log_location_file):
    write_query_log(
        "123",
        10,
        0,
        timedelta(minutes=2),
        True,
        path_to_log_dir=log_location_file.parent,
    )
    assert log_location_file.exists()
    t = log_location_file.read_text().split("\n")[1]
    assert t == "123,10,0,120.0,1"


def test_log_adaptor(log_location_file):
    log_adaptor.write_query_log(
        "123",
        10,
        0,
        timedelta(seconds=20),
        True,
        path_to_log_dir=log_location_file.parent,
    )
    assert log_location_file.exists()
    t = log_location_file.read_text().split("\n")[1]
    assert t == "123,10,0,20.0,1"


def test_log_file_write_2lines(log_location_file):
    write_query_log(
        "123",
        10,
        0,
        timedelta(minutes=2),
        True,
        path_to_log_dir=log_location_file.parent,
    )
    write_query_log(
        "124",
        10,
        0,
        timedelta(minutes=2),
        True,
        path_to_log_dir=log_location_file.parent,
    )
    assert log_location_file.exists()
    t = log_location_file.read_text()
    assert len(t.split("\n")) == 4


def test_callback_no_updates():
    called = False

    def call_me(total: Optional[int], processed: int, downloaded: int, failed: int):
        nonlocal called
        called = True

    u = _status_update_wrapper(call_me)
    u.broadcast()


def test_callback_processed_set():
    "Sometimes Servicex gets called when there is no total yet."

    called = False
    p_total = None
    p_processed = None
    p_downloaded = None
    p_failed = None

    def call_me(total: Optional[int], processed: int, downloaded: int, failed: int):
        nonlocal called, p_total, p_processed, p_downloaded, p_failed
        called = True
        p_total = total
        p_processed = processed
        p_downloaded = downloaded
        p_failed = failed

    u = _status_update_wrapper(call_me)
    u.update(processed=10)
    u.broadcast()
    assert called
    assert p_total is None
    assert p_downloaded == 0
    assert p_processed == 10
    assert p_failed == 0


def test_callback_processed_and_downloaded():
    "Sometimes Servicex gets called when there is no total yet."

    called = False
    p_total = None
    p_processed = None
    p_downloaded = None
    p_failed = None

    def call_me(total: Optional[int], processed: int, downloaded: int, failed: int):
        nonlocal called, p_total, p_processed, p_downloaded, p_failed
        called = True
        p_total = total
        p_processed = processed
        p_downloaded = downloaded
        p_failed = failed

    u = _status_update_wrapper(call_me)
    u.update(processed=10)
    u.update(downloaded=3)
    u.broadcast()
    assert called
    assert p_total is None
    assert p_downloaded == 3
    assert p_processed == 10
    assert p_failed == 0


def test_callback_everything():
    "Sometimes Servicex gets called when there is no total yet."

    called = False
    p_total = None
    p_processed = None
    p_downloaded = None
    p_failed = None

    def call_me(total: Optional[int], processed: int, downloaded: int, failed: int):
        nonlocal called, p_total, p_processed, p_downloaded, p_failed
        called = True
        p_total = total
        p_processed = processed
        p_downloaded = downloaded
        p_failed = failed

    u = _status_update_wrapper(call_me)
    u.update(processed=10)
    u.update(downloaded=2)
    u.update(failed=1)
    u.update(remaining=1)
    u.broadcast()
    assert called
    assert p_total == 12
    assert p_downloaded == 2
    assert p_processed == 10
    assert p_failed == 1


def test_callback_reset():
    "We want to restart."

    called = False
    p_total = None
    p_processed = None
    p_downloaded = None
    p_failed = None

    def call_me(total: Optional[int], processed: int, downloaded: int, failed: int):
        nonlocal called, p_total, p_processed, p_downloaded, p_failed
        called = True
        p_total = total
        p_processed = processed
        p_downloaded = downloaded
        p_failed = failed

    u = _status_update_wrapper(call_me)
    u.update(processed=10)
    u.update(downloaded=2)
    u.update(failed=1)
    u.update(remaining=2)
    u.reset()
    u.update(processed=10)
    u.update(downloaded=2)
    u.update(remaining=2)
    u.broadcast()
    assert called
    assert p_total == 12
    assert p_downloaded == 2
    assert p_processed == 10
    assert p_failed == 0


def test_callback_inc_with_nothing():
    called = False
    p_total = None
    p_processed = None
    p_downloaded = None
    p_failed = None

    def call_me(total: Optional[int], processed: int, downloaded: int, failed: int):
        nonlocal called, p_total, p_processed, p_downloaded, p_failed
        called = True
        p_total = total
        p_processed = processed
        p_downloaded = downloaded
        p_failed = failed

    u = _status_update_wrapper(call_me)
    u.inc(downloaded=3)
    u.broadcast()
    assert p_downloaded == 3


def test_callback_inc_with_already_set():
    called = False
    p_total = None
    p_processed = None
    p_downloaded = None
    p_failed = None

    def call_me(total: Optional[int], processed: int, downloaded: int, failed: int):
        nonlocal called, p_total, p_processed, p_downloaded, p_failed
        called = True
        p_total = total
        p_processed = processed
        p_downloaded = downloaded
        p_failed = failed

    u = _status_update_wrapper(call_me)
    u.update(downloaded=1)
    u.inc(downloaded=3)
    u.broadcast()
    assert p_downloaded == 4


def test_callback_none():
    "What if there is no callback?"

    u = _status_update_wrapper(None)
    u.update(processed=10)
    u.update(downloaded=3)
    u.update(remaining=12)
    u.update(failed=1)
    u.broadcast()


def test_callback_with_total_fluctuation():
    "Sometimes we get the total wrong.."

    p_total = None

    def call_me(total: Optional[int], processed: int, downloaded: int, failed: int):
        nonlocal p_total
        p_total = total

    u = _status_update_wrapper(call_me)
    u.update(processed=6, remaining=6)
    u.broadcast()
    assert p_total == 12
    u.update(processed=6, remaining=5)
    u.broadcast()
    assert p_total == 12
    u.update(processed=7, remaining=6)
    u.broadcast()
    assert p_total == 13


def test_callback_with_total_sequence():
    "Make sure we can del with multiple things at once"

    p_total = None

    def call_me(total: Optional[int], processed: int, downloaded: int, failed: int):
        nonlocal p_total
        p_total = total

    u = _status_update_wrapper(call_me)
    u.update(processed=0, remaining=6)
    u.update(processed=2, remaining=4)
    u.broadcast()
    assert p_total == 6


@pytest.mark.asyncio
async def test_transform_sequence():
    u = _status_update_wrapper()

    v = [
        i async for i in stream_status_updates(as_async_seq([(1, 0, 0), (0, 1, 0)]), u)
    ]

    assert len(v) == 2
    assert u.failed == 0
    assert u.total == 1


@pytest.mark.asyncio
async def test_transform_updates_unique():

    v = [
        i
        async for i in stream_unique_updates_only(as_async_seq([(1, 0, 0), (0, 1, 0)]))
    ]

    assert len(v) == 2


@pytest.mark.asyncio
async def test_transform_updates_unique_2():

    v = [
        i
        async for i in stream_unique_updates_only(as_async_seq([(1, 0, 0), (1, 0, 0)]))
    ]

    assert len(v) == 1


def test_cache_stable():
    json_query = {
        "did": "dude_001",
        "selection": "(valid qastle)",
        "chunk-size": 1000,
        "workers": 10,
    }
    h1 = _query_cache_hash(json_query)
    h2 = _query_cache_hash(json_query)

    assert h1 == h2


def test_cache_query_agnostic():
    json_query = {
        "did": "dude_001",
        "selection": "(lambda (list a0 a1) (+ a0 a1))",
        "chunk-size": 1000,
        "workers": 10,
    }
    h1 = _query_cache_hash(json_query)
    json_query["selection"] = "(lambda (list b0 b1) (+ b0 b1))"
    h2 = _query_cache_hash(json_query)

    assert h1 == h2


def test_request_trans_cache_workers():
    json_query = {
        "did": "dude_001",
        "selection": "(valid qastle)",
        "chunk-size": 1000,
        "workers": 10,
    }
    h1 = _query_cache_hash(json_query)
    json_query["workers"] = 100
    h2 = _query_cache_hash(json_query)
    assert h1 == h2


def test_request_trans_cache_selection():
    json_query = {
        "did": "dude_001",
        "selection": "(valid qastle)",
        "chunk-size": 1000,
        "workers": 10,
    }
    h1 = _query_cache_hash(json_query)
    json_query["selection"] = "(call valid qastle)"
    h2 = _query_cache_hash(json_query)
    assert h1 != h2


def test_request_trans_cache_did():
    json_query = {
        "did": "dude_001",
        "selection": "(valid qastle)",
        "chunk-size": 1000,
        "workers": 10,
    }
    h1 = _query_cache_hash(json_query)
    json_query["did"] = "did_002"
    h2 = _query_cache_hash(json_query)
    assert h1 != h2


def test_request_trans_cache_unknown():
    json_query = {
        "did": "dude_001",
        "selection": "(valid qastle)",
        "chunk-size": 1000,
        "workers": 10,
    }
    h1 = _query_cache_hash(json_query)
    json_query["fork"] = "me"
    h2 = _query_cache_hash(json_query)
    assert h1 != h2


def test_clean_linq_no_lambda():
    q = "(valid qastle)"
    assert clean_linq(q) == q


def test_clean_linq_bad_lambda():
    q = "(lambda a (+ a 1) (1 2 3))"
    assert clean_linq(q) == q


def test_clean_linq_invalid_qastle():
    q = "(lambda a (+ a 1))"
    assert clean_linq(q) == q


def test_clean_linq_single_lambda():
    q = "(lambda (list e0 e1) (+ e0 e1))"
    assert clean_linq(q) == "(lambda (list a0 a1) (+ a0 a1))"


def test_clean_linq_many_args():
    q = (
        "(lambda (list e0 e1 e2 e3 e4 e5 e6 e7 e8 e9 e10 e11) "
        "(+ e0 e1 e2 e3 e4 e5 e6 e7 e8 e9 e10 e11))"
    )
    assert (
        clean_linq(q) == "(lambda (list a0 a1 a2 a3 a4 a5 a6 a7 a8 a9 a10 a11) "
        "(+ a0 a1 a2 a3 a4 a5 a6 a7 a8 a9 a10 a11))"
    )


def test_clean_linq_funny_var_names():
    q = "(lambda (list e0 e000) (+ e000 e0))"
    assert clean_linq(q) == "(lambda (list a0 a1) (+ a1 a0))"


def test_clean_linq_nested_lambda():
    q = "(lambda (list e0 e1) (+ (call (lambda (list e0) (+ e0 1)) e0) e1))"
    assert (
        clean_linq(q)
        == "(lambda (list a1 a2) (+ (call (lambda (list a0) (+ a0 1)) a1) a2))"
    )


def test_clean_linq_arb_var_names():
    q = "(lambda (list a0 f1) (+ a0 f1))"
    assert clean_linq(q) == "(lambda (list a0 a1) (+ a0 a1))"


def test_clean_linq_empty():
    q = ""
    assert clean_linq(q) == ""


def test_configured_cache_default():
    from confuse import Configuration

    c = Configuration("servicex", "servicex")
    assert c["cache_path"].exists()


def test_configured_cache_temp_location():
    from confuse import Configuration

    c = Configuration("bogus", "bogus")
    c.clear()
    c["cache_path"] = "/tmp/servicex-dude"

    p = get_configured_cache_path(c)

    # Should default to temp directory - should work on all platforms!
    assert p.exists()
    assert str(p).startswith(tempfile.gettempdir())
    assert "servicex-dude" in str(p)


def test_configured_cache_location():
    from confuse import Configuration

    c = Configuration("bogus", "bogus")
    c.clear()
    here = Path("./servicex-dude").absolute()
    c["cache_path"] = str(here)

    p = get_configured_cache_path(c)

    # Should default to temp directory - should work on all platforms!
    assert p.exists()
    assert str(p) == str(here)


def test_cache_expansion_user():
    """On windows this will expand one way, and on linux another. So need to be a little careful here!"""
    # If we are running in a place where USER/Username does not exist!
    if "USER" not in os.environ and "UserName" not in os.environ:
        os.environ["USER"] = "bogus"

    from confuse import Configuration

    c = Configuration("bogus", "bogus")
    c.clear()
    c["cache_path"] = "/tmp/servicex_${USER}"

    # Get the right answer, depending on the definition of USER
    u_name = os.environ["USER"] if "USER" in os.environ else os.environ["UserName"]
    path_name = f"servicex_{u_name}"

    p = get_configured_cache_path(c)

    # Should default to temp directory - should work on all platforms!
    assert p.name == path_name


def test_cache_expansion_username():
    """On windows this will expand one way, and on linux another. So need to be a little careful here!"""
    # If we are running in a place where USER/Username does not exist!
    if "USER" not in os.environ and "UserName" not in os.environ:
        os.environ["USER"] = "bogus"

    from confuse import Configuration

    c = Configuration("bogus", "bogus")
    c.clear()
    c["cache_path"] = "/tmp/servicex_${UserName}"

    # Get the right answer, depending on the definition of USER
    u_name = os.environ["USER"] if "USER" in os.environ else os.environ["UserName"]
    path_name = f"servicex_{u_name}"

    p = get_configured_cache_path(c)

    # Should default to temp directory - should work on all platforms!
    assert p.name == path_name


def test_bar_name_title():
    "Test various uses of the progress bar title generator"
    assert dataset_as_name(None, None) == "<none>"

    assert dataset_as_name("sample1", None) == "sample1"
    assert (
        dataset_as_name("sample1sample1sample1sample1", None)
        == "sample1sample1sample..."
    )

    assert dataset_as_name(["s1", "s2", "s3"], None) == "[s1, s2, s3]"
    assert (
        dataset_as_name(
            ["sample1sample1sample1sample1", "sample1sample1sample1sample1"], None
        )
        == "[sample1sample1sampl..."
    )

    assert dataset_as_name(sample_name="sample1", title="hi") == "hi"
    assert (
        dataset_as_name("hi", "sample1sample1sample1sample1")
        == "sample1sample1sample..."
    )


def test_bar_name_long():
    "Try a longer version"
    assert (
        dataset_as_name("sample1sample1sample1sample1", max_len=None)
        == "sample1sample1sample1sample1"
    )


@pytest.mark.asyncio()
async def test_async_itr_good():
    @on_exception_itr(backoff.constant, ValueError, interval=0.01, max_tries=2)
    async def return_two():
        yield 1
        yield 2

    lst = [item async for item in return_two()]
    assert lst == [1, 2]


@pytest.mark.asyncio()
async def test_async_itr_exception_first():
    first = True

    @on_exception_itr(backoff.constant, ValueError, interval=0.01, max_tries=2)
    async def return_two():
        nonlocal first
        if first:
            first = False
            raise ValueError("hi")
        yield 1
        yield 2

    lst = [item async for item in return_two()]
    assert lst == [1, 2]


@pytest.mark.asyncio()
async def test_async_itr_exception_not_first():
    count = 0

    @on_exception_itr(backoff.constant, ValueError, interval=0.01, max_tries=2)
    async def return_two():
        nonlocal count
        count = count + 1
        yield 1
        raise ValueError("hi")

    with pytest.raises(ValueError) as e:
        [item async for item in return_two()]

    assert count == 1
    assert "hi" in str(e)


@pytest.mark.asyncio()
async def test_async_itr_exception_all():
    count = 0

    @on_exception_itr(backoff.constant, ValueError, interval=0.01, max_tries=2)
    async def return_two():
        nonlocal count
        count = count + 1
        raise ValueError("hi")
        yield 1  # NOQA

    with pytest.raises(ValueError) as e:
        [item async for item in return_two()]

    assert count == 2
    assert "hi" in str(e)


@pytest.mark.asyncio()
async def test_async_itr_exception_all_wait_iter():
    count = 0

    @on_exception_itr(backoff.constant, ValueError, interval=0.01, max_tries=2)
    async def return_two():
        nonlocal count
        count = count + 1
        raise ValueError("hi")
        yield 1

    with pytest.raises(ValueError) as e:
        [item async for item in return_two()]

    assert count == 2
    assert "hi" in str(e)



import random
from servicex.cache import reset_local_query_cache
import string
from pathlib import Path

import pytest
from servicex import Cache, ServiceXException, ignore_cache, update_local_query_cache


@pytest.fixture()
def reset_in_memory_cache():
    Cache.reset_cache()
    yield
    Cache.reset_cache()


@pytest.fixture(autouse=True)
def analysis_query_cache_reset():
    "Want to make sure the local query cache is reset for everything!"
    reset_local_query_cache()
    cf = Path("./servicex_query_cache.json")
    if cf.exists():
        cf.unlink()
    yield
    reset_local_query_cache()
    if cf.exists():
        cf.unlink()


def test_create_cache(tmp_path):
    c = Cache(tmp_path)
    assert c.path == tmp_path


def test_query_miss(tmp_path):
    c = Cache(tmp_path)

    assert c.lookup_query({"hi": "there"}) is None


def test_query_hit_1(tmp_path):
    c = Cache(tmp_path)
    c.set_query({"hi": "there"}, "dude")
    assert c.lookup_query({"hi": "there"}) == "dude"


def test_analysis_cache_set_twice_different(tmp_path: Path):
    "Make sure we cannot set with two different paths"
    update_local_query_cache(tmp_path / "analysis_cache1.json")
    with pytest.raises(ServiceXException):
        update_local_query_cache(tmp_path / "analysis_cache2.json")


def test_analysis_cache_set_twice_different_paths(tmp_path: Path):
    "Make sure we cannot set with two different paths"
    update_local_query_cache(tmp_path / "bogus1" / "analysis_cache.json")
    with pytest.raises(ServiceXException):
        update_local_query_cache(tmp_path / "bogus2" / "analysis_cache.json")


def test_analysis_cache_set_twice_same(tmp_path: Path):
    "Make sure we can set with two same paths"
    update_local_query_cache(tmp_path / "analysis_cache1.json")
    update_local_query_cache(tmp_path / "analysis_cache1.json")


def test_query_hit_analysis_cache(tmp_path: Path):
    "Make sure the analysis cache is updated"
    cache_loc_1 = tmp_path / "cache1"
    cache_loc_2 = tmp_path / "cache2"

    update_local_query_cache(tmp_path / "analysis_cache.json")

    c1 = Cache(cache_loc_1)
    c1.set_query({"hi": "there"}, "dude")

    c2 = Cache(cache_loc_2)
    assert c2.lookup_query({"hi": "there"}) == "dude"


def test_query_hit_analysis_no_cache(tmp_path: Path):
    "Make sure no caching occurs if the user doesn't want it to"
    cache_loc_1 = tmp_path / "cache1"
    cache_loc_2 = tmp_path / "cache2"

    c1 = Cache(cache_loc_1)
    c1.set_query({"hi": "there"}, "dude")

    c2 = Cache(cache_loc_2)
    assert c2.lookup_query({"hi": "there"}) is None


def test_query_miss_analysis_cache_keys(tmp_path: Path):
    "Make sure the analysis cache is updated"
    cache_loc_1 = tmp_path / "cache1"
    cache_loc_2 = tmp_path / "cache2"

    update_local_query_cache(tmp_path / "analysis_cache.json")

    c1 = Cache(cache_loc_1)
    c1.set_query({"hi": "there"}, "dude")

    c2 = Cache(cache_loc_2, analysis_query_key="bogus")
    assert c2.lookup_query({"hi": "there"}) is None


def test_query_hit_analysis_cache_update(tmp_path: Path):
    "Make sure the analysis cache is updated"
    cache_loc_1 = tmp_path / "cache1"
    cache_loc_2 = tmp_path / "cache2"

    update_local_query_cache(tmp_path / "analysis_cache.json")

    c1 = Cache(cache_loc_1)
    c1.set_query({"hi": "there"}, "dude")
    c1.set_query({"hi1": "there1"}, "dude1")

    c2 = Cache(cache_loc_2)
    assert c2.lookup_query({"hi": "there"}) == "dude"
    assert c2.lookup_query({"hi1": "there1"}) == "dude1"


def test_query_hit_analysis_cache_silent_read(tmp_path: Path):
    "Make sure we read the analysis cache if it happens to be present"
    cache_loc_1 = tmp_path / "cache1"
    cache_loc_2 = tmp_path / "cache2"

    update_local_query_cache()

    c1 = Cache(cache_loc_1)
    c1.set_query({"hi": "there"}, "dude")

    reset_local_query_cache()

    c2 = Cache(cache_loc_2)
    assert c2.lookup_query({"hi": "there"}) == "dude"


def test_query_hit_analysis_cache_empty_file(tmp_path: Path):
    "Make sure we read the analysis cache if it happens to be present"
    cache_loc_1 = tmp_path / "cache1"
    cache_loc_2 = tmp_path / "cache2"

    qc = Path("servicex_query_cache.json")
    qc.touch()

    c1 = Cache(cache_loc_1)
    c1.set_query({"hi": "there"}, "dude")

    reset_local_query_cache()

    c2 = Cache(cache_loc_2)
    assert c2.lookup_query({"hi": "there"}) == "dude"


def test_query_hit_analysis_cache_not_triggered(tmp_path: Path):
    "Make sure analysis cache is not written by default"
    cache_loc_1 = tmp_path / "cache1"
    cache_loc_2 = tmp_path / "cache2"

    c1 = Cache(cache_loc_1)
    c1.set_query({"hi": "there"}, "dude")

    c2 = Cache(cache_loc_2)
    assert c2.lookup_query({"hi": "there"}) is None


def test_query_hit_analysis_lookup_writes(tmp_path: Path):
    "make sure analysis query cache is written to"
    cache_loc_1 = tmp_path / "cache1"
    cache_loc_2 = tmp_path / "cache2"

    c1 = Cache(cache_loc_1)
    c1.set_query({"hi": "there"}, "dude")
    update_local_query_cache(tmp_path / "analysis_cache.json")
    assert c1.lookup_query({"hi": "there"}) == "dude"

    c2 = Cache(cache_loc_2)
    assert c2.lookup_query({"hi": "there"}) == "dude"


def test_query_hit_analysis_cache_removed_query_no_update(tmp_path: Path):
    "Make sure to forget a query when we are not updating the analysis cache"
    cache_loc_1 = tmp_path / "cache1"
    cache_loc_2 = tmp_path / "cache2"
    cache_loc_3 = tmp_path / "cache3"

    update_local_query_cache()

    c1 = Cache(cache_loc_1)
    c1.set_query({"hi": "there"}, "dude")

    reset_local_query_cache()

    c2 = Cache(cache_loc_2)
    c2.remove_query({"hi": "there"})
    assert c2.lookup_query({"hi": "there"}) is None

    reset_local_query_cache()

    # Make sure that the json file wasn't modified in this case!
    c3 = Cache(cache_loc_3)
    assert c3.lookup_query({"hi": "there"}) == "dude"


def test_query_hit_analysis_cache_removed_query_update(tmp_path: Path):
    "If we are updating the query cache, make sure to remove an item"
    cache_loc_1 = tmp_path / "cache1"
    cache_loc_2 = tmp_path / "cache2"

    update_local_query_cache()

    c1 = Cache(cache_loc_1)
    c1.set_query({"hi": "there"}, "dude")
    c1.remove_query({"hi": "there"})

    c2 = Cache(cache_loc_2)
    assert c2.lookup_query({"hi": "there"}) is None


def test_ic_query(tmp_path):
    c = Cache(tmp_path)
    c.set_query({"hi": "there"}, "dude")
    with ignore_cache():
        assert c.lookup_query({"hi": "there"}) is None


def test_ic_query_query_context(tmp_path):
    c = Cache(tmp_path)
    c.set_query({"hi": "there"}, "dude")
    with c.ignore_cache():
        assert c.lookup_query({"hi": "there"}) is None


def test_ic_query_ds_level(tmp_path):
    c = Cache(tmp_path, ignore_cache=True)
    c.set_query({"hi": "there"}, "dude")
    assert c.lookup_query({"hi": "there"}) is None


def test_query_hit_2(tmp_path):
    c = Cache(tmp_path)
    c.set_query({"hi": "there"}, "dude1")
    c.set_query({"hi": "there_1"}, "dude2")
    assert c.lookup_query({"hi": "there"}) == "dude1"
    assert c.lookup_query({"hi": "there_1"}) == "dude2"


def test_query_lookup_from_file(tmp_path):
    c1 = Cache(tmp_path)
    c1.set_query({"hi": "there"}, "dude")

    c2 = Cache(tmp_path)
    assert c2.lookup_query({"hi": "there"}) == "dude"


def test_query_remove(tmp_path):
    c1 = Cache(tmp_path)
    c1.set_query({"hi": "there"}, "dude")
    c1.remove_query({"hi": "there"})
    assert c1.lookup_query({"hi": "there"}) is None


def test_files_miss(tmp_path):
    c = Cache(tmp_path)
    assert c.lookup_files("1234") is None


def test_files_hit(tmp_path: Path):
    c = Cache(tmp_path)
    f1 = tmp_path / "f1.root"
    f2 = tmp_path / "f2.root"
    f1.touch()
    f2.touch()
    c.set_files("1234", [("hi", f1), ("there", f2)])
    assert c.lookup_files("1234") == [("hi", f1), ("there", f2)]


def test_files_deleted_data_file(tmp_path: Path):
    c = Cache(tmp_path)
    f1 = tmp_path / "f1.root"
    f2 = tmp_path / "f2.root"
    f1.touch()
    f2.touch()
    c.set_files("1234", [("hi", f1), ("there", f2)])
    f2.unlink()
    assert c.lookup_files("1234") is None


def test_ic_files_hit(tmp_path: Path):
    "The file list should not be affected by cache ignores"
    c = Cache(tmp_path)
    f1 = tmp_path / "f1.root"
    f2 = tmp_path / "f2.root"
    f1.touch()
    f2.touch()
    c.set_files("1234", [("hi", f1), ("there", f2)])
    with ignore_cache():
        assert c.lookup_files("1234") == [("hi", f1), ("there", f2)]


def test_files_hit_reloaded(tmp_path):
    c1 = Cache(tmp_path)
    f1 = tmp_path / "f1.root"
    f2 = tmp_path / "f2.root"
    f1.touch()
    f2.touch()
    c1.set_files("1234", [("hi", f1), ("there", f2)])
    c2 = Cache(tmp_path)
    assert c2.lookup_files("1234") == [("hi", f1), ("there", f2)]


def test_memory_miss(tmp_path):
    c = Cache(tmp_path)
    assert c.lookup_inmem("dude") is None


def test_memory_hit(tmp_path):
    c = Cache(tmp_path)
    r = 10
    c.set_inmem("dude", r)
    assert c.lookup_inmem("dude") is r


def test_ic_memory_hit(tmp_path):
    c = Cache(tmp_path)
    r = 10
    c.set_inmem("dude", r)
    with ignore_cache():
        assert c.lookup_inmem("dude") is None


def test_ic_memory_hit_ds_context(tmp_path):
    c = Cache(tmp_path)
    r = 10
    c.set_inmem("dude", r)
    with c.ignore_cache():
        assert c.lookup_inmem("dude") is None


def test_ic_memory_hit_ds_level(tmp_path):
    c = Cache(tmp_path, ignore_cache=True)
    r = 10
    c.set_inmem("dude", r)
    assert c.lookup_inmem("dude") is None


def test_memory_hit_accross(tmp_path):
    c1 = Cache(tmp_path)
    r = 10
    c1.set_inmem("dude", r)
    c2 = Cache(tmp_path)
    assert c2.lookup_inmem("dude") is r


def test_data_file_location(tmp_path):
    c = Cache(tmp_path)
    p = c.data_file_location("123-456", "junk.root")
    assert not p.exists()
    p.touch()
    assert p.exists()
    assert str(p).startswith(str(tmp_path))


def test_data_file_location_long_path(tmp_path):
    c = Cache(tmp_path)
    letters = string.ascii_lowercase
    file_significant_name = "junk.root"
    long_file_name = "".join(random.choice(letters) for i in range(230))

    p = c.data_file_location("123-456", long_file_name + file_significant_name)
    assert len(p.name) == c.max_path_len - len(p.parent.name)
    assert p.name.endswith(file_significant_name)


def test_data_file_location_twice(tmp_path):
    c = Cache(tmp_path)
    _ = c.data_file_location("123-456", "junk1.root")
    p2 = c.data_file_location("123-456", "junk2.root")
    assert not p2.exists()
    p2.touch()
    assert p2.exists()


def test_data_file_bad_file(tmp_path):
    "Check a very long bad filename to make sure it is sanitized"
    c = Cache(tmp_path)
    p = c.data_file_location(
        "123-456",
        "root:::dcache-atlas-xrootd-wan.desy.de:1094::pnfs:desy.de:atlas"
        ":dq2:atlaslocalgroupdisk:rucio:mc15_13TeV:8a:f1:DAOD_STDM3.05630052"
        "._000001.pool.root.198fbd841d0a28cb0d9dfa6340c890273-1.part.minio",
    )
    assert not p.exists()
    # If the follow fails, on windows, it could be because very-long-pathnames are not
    # enabled.
    p.touch()
    assert p.exists()


def test_query_cache_status(tmp_path):
    c = Cache(tmp_path)

    info = {"request_id": "111-222-333", "key": "bogus"}
    c.set_query_status(info)
    assert c.query_status_exists("111-222-333")
    info1 = c.lookup_query_status("111-222-333")
    assert info1["key"] == "bogus"


def test_query_cache_status_bad(tmp_path):
    c = Cache(tmp_path)

    assert not c.query_status_exists("111-222-333")
    with pytest.raises(ServiceXException):
        c.lookup_query_status("111-222-333")


def test_ic_query_cache_status(tmp_path):
    "Query status should be cached and accessed *during* a query"
    c = Cache(tmp_path)
    info = {"request_id": "111-222-333", "key": "bogus"}
    c.set_query_status(info)
    with ignore_cache():
        info1 = c.lookup_query_status("111-222-333")
        assert info1["key"] == "bogus"


def test_ic_restore(tmp_path):
    c = Cache(tmp_path)
    c.set_query({"hi": "there"}, "dude")
    with ignore_cache():
        pass
    assert c.lookup_query({"hi": "there"}) == "dude"


def test_ic_nesting(tmp_path):
    c = Cache(tmp_path)
    c.set_query({"hi": "there"}, "dude")
    with ignore_cache():
        with ignore_cache():
            pass
        assert c.lookup_query({"hi": "there"}) is None


def test_ic_nesting_ds_context(tmp_path):
    c = Cache(tmp_path)
    c.set_query({"hi": "there"}, "dude")
    with c.ignore_cache():
        with c.ignore_cache():
            pass
        assert c.lookup_query({"hi": "there"}) is None


def test_ic_enter_exit(tmp_path):
    c = Cache(tmp_path)
    c.set_query({"hi": "there"}, "dude")
    i = ignore_cache()
    i.__enter__()
    assert c.lookup_query({"hi": "there"}) is None
    i.__exit__(None, None, None)
    assert c.lookup_query({"hi": "there"}) == "dude"



import asyncio
from datetime import datetime
from typing import AsyncIterator, Dict, Optional, Tuple
import logging

import aiohttp
from google.auth import jwt

from .utils import (
    ServiceXException,
    ServiceXFailedFileTransform,
    ServiceXFatalTransformException,
    ServiceXUnknownRequestID,
    TransformTuple,
)

# Number of seconds to wait between polling servicex for the status of a transform job
# while waiting for it to finish.
servicex_status_poll_time = 5.0


# Low level routines for interacting with a ServiceX instance via the WebAPI
class ServiceXAdaptor:
    def __init__(self, endpoint, refresh_token=None):
        """
        Authenticated access to ServiceX
        """
        self._endpoint = endpoint
        self._token = None
        self._refresh_token = refresh_token

    async def _get_token(self, client: aiohttp.ClientSession):
        url = f"{self._endpoint}/token/refresh"
        headers = {"Authorization": f"Bearer {self._refresh_token}"}
        async with client.post(url, headers=headers, json=None) as response:
            status = response.status
            if status == 200:
                j = await response.json()
                self._token = j["access_token"]
            else:
                raise ServiceXException(
                    f"ServiceX access token request rejected: {status}"
                )

    async def _get_authorization(self, client: aiohttp.ClientSession):
        if not self._refresh_token:
            return {}
        now = datetime.utcnow().timestamp()
        if not self._token or jwt.decode(self._token, verify=False)["exp"] - now < 0:
            await self._get_token(client)
        return {"Authorization": f"Bearer {self._token}"}

    async def submit_query(
        self, client: aiohttp.ClientSession, json_query: Dict[str, str]
    ) -> Dict[str, str]:
        """
        Submit a query to ServiceX, and return a request ID
        """

        headers = await self._get_authorization(client)

        async with client.post(
            f"{self._endpoint}/servicex/transformation",
            headers=headers,
            json=json_query,
        ) as response:
            status = response.status
            if status != 200:
                # This was an error at ServiceX, bubble it up so code above us can
                # handle as needed.
                t = await response.text()
                raise ServiceXException(
                    "ServiceX rejected the transformation request: " f"({status}){t}"
                )

            r = await response.json()
            return r

    async def get_query_status(
        self, client: aiohttp.ClientSession, request_id: str
    ) -> Dict[str, str]:
        """Returns the full query information from the endpoint.

        Args:
            client (aiohttp.ClientSession): Client session on which to make the request.
            request_id (str): The request id to return the tranform status

        Raises:
            ServiceXException: If we fail to find the information.

        Returns:
            Dict[str, str]: The JSON dictionary of information returned from ServiceX
        """
        headers = await self._get_authorization(client)

        async with client.get(
            f"{self._endpoint}/servicex/transformation/{request_id}", headers=headers
        ) as response:
            status = response.status
            if status != 200:
                # This was an error at ServiceX, bubble it up so code above us can
                # handle as needed.
                t = await response.text()
                raise ServiceXException(
                    "ServiceX rejected the transformation status fetch: "
                    f"({status}){t}"
                )

            r = await response.json()
            return r

    async def dump_query_errors(self, client: aiohttp.ClientSession, request_id: str):
        """Dumps to the logging system any error messages we find from ServiceX.

        Args:
            client (aiohttp.ClientSession): Client along which to send queries.
            request_id (str): Fetch all errors from there.
        """

        headers = await self._get_authorization(client)
        async with client.get(
            f"{self._endpoint}/servicex/transformation/{request_id}/errors",
            headers=headers,
        ) as response:
            status = response.status
            if status != 200:
                t = await response.text()
                if "Request not found" in t:
                    raise ServiceXUnknownRequestID(
                        f"Unable to get errors for request {request_id}"
                        f": {status} - {t}"
                    )
                else:
                    raise ServiceXException(
                        f"Failed to get request errors for {request_id}: "
                        f"{status} - {t}"
                    )

            # Dump the messages out to the logger if there are any!
            # Keep the output under control - only dump the first 20 errors (see #188)
            errors = (await response.json())["errors"]
            log = logging.getLogger(__name__)
            log.warning(f"Transform {request_id} had {len(errors)} errors:")
            for e in errors[:20]:
                log.warning(f'  Error transforming file: {e["file"]}')
                for ln in e["info"].split("\n"):
                    log.warning(f"  -> {ln}")

    @staticmethod
    def _get_transform_stat(info: Dict[str, str], stat_name: str) -> Optional[int]:
        "Return the info from a servicex status reply, protecting against bad internet returns"
        return (
            None
            if ((stat_name not in info) or (info[stat_name] is None))
            else int(info[stat_name])
        )

    async def get_transform_status(
        self, client: aiohttp.ClientSession, request_id: str
    ) -> Tuple[Optional[int], int, Optional[int]]:
        """
        Internal routine that queries for the current stat of things. We expect the
        following things to come back:
            - files-processed
            - files-remaining
            - files-skipped
            - request-id
            - stats

        If the transform has already completed, we return data from cache.

        Arguments:

            endpoint            Web API address where servicex lives
            request_id          The id of the request to check up on

        Raises:

            ServiceXException   If the status returns `Fatal`.

        Returns:

            files_remaining     How many files remain to be processed. None if the number
                                has not yet been determined
            files_processed     How many files have been successfully processed
                                by the system.
            files_failed        Number of files that were skipped
        """
        headers = await self._get_authorization(client)

        # Make the actual query
        async with client.get(
            f"{self._endpoint}/servicex/transformation/{request_id}/status",
            headers=headers,
        ) as response:
            status = response.status
            if status != 200:
                raise ServiceXUnknownRequestID(
                    f"Unable to get transform status "
                    f"for request id {request_id}"
                    f" - http error {status}"
                )
            info = await response.json()
            logging.getLogger(__name__).debug(
                f"Status response for {request_id}: {info}"
            )

            if "status" in info and info["status"] == "Fatal":
                raise ServiceXFatalTransformException(
                    f"Transform status for {request_id}" ' is marked "Fatal".'
                )

            files_remaining = self._get_transform_stat(info, "files-remaining")
            files_failed = self._get_transform_stat(info, "files-skipped")
            files_processed = self._get_transform_stat(info, "files-processed")

            if files_remaining is not None and int(files_remaining) < 0:
                raise ServiceXFatalTransformException(
                    f"Transform {request_id} returned negative files remaining. Bug in ServiceX?"
                )

            assert files_processed is not None

            return files_remaining, files_processed, files_failed


async def transform_status_stream(
    sa: ServiceXAdaptor, client: aiohttp.ClientSession, request_id: str
) -> AsyncIterator[TransformTuple]:
    """
    Returns an async stream of `(files-remaining, files_processed, files_failed)` until the
    servicex `request_id` request is finished, against the servicex instance located at
    `sa`.

    Arguments:

        sa                  The servicex low level adaptor
        client              An async http function we can call and use
        request_id          The request id for this request

    Returns:

        remaining, processed, skipped     Returns an async stream triple of the
                                          status numbers. Every time we find something
                                          we send it on.

    Note:
    """
    done = False
    while not done:
        next_processed = await sa.get_transform_status(client, request_id)
        remaining, _, _ = next_processed
        done = remaining is not None and remaining == 0
        yield next_processed

        if not done:
            await asyncio.sleep(servicex_status_poll_time)


async def trap_servicex_failures(
    stream: AsyncIterator[TransformTuple],
) -> AsyncIterator[TransformTuple]:
    """
    Looks for any failed files. If it catches one, it will remember it and throw once the stream
    is done. This allows all the files to come down first.
    """
    async for p in stream:
        remain, processed, did_fail = p
        if did_fail is not None and did_fail != 0:
            raise ServiceXFailedFileTransform(
                f"ServiceX failed to transform {did_fail} "
                f"files - data incomplete (remaining: {remain}, "
                f"processed: {processed})."
            )

        yield p



import asyncio
from typing import Dict
import functools
import logging

from .utils import _string_hash, clean_linq

_in_progress_items: Dict[str, asyncio.Event] = {}


def _wrap_in_memory_sx_cache(fn):
    """
    Caching wrapper. This is like lru in python, but we have to fetch some more
    "interesting" data from the object we are wrapping to complete the cache key. Hence
    this implementation.

    NOTE: This is not thread safe. In fact, you'll likely get a bomb if try to
          call this from another thread and the two threads request something the
          same at the same time. This is because asyncio locks are associated with loops,
          and there is one loop per thread.
    """

    @functools.wraps(fn)
    async def cached_version_of_fn(*args, **kwargs):
        assert len(args) == 2 or len(args) == 3
        sx = args[0]
        from .servicex import ServiceXDataset

        assert isinstance(sx, ServiceXDataset)
        selection_query = args[1]
        assert isinstance(selection_query, str)

        # Is it in the local cache?
        h = _string_hash([sx._dataset, fn.__name__, clean_linq(selection_query)])
        logger = logging.getLogger(__name__)
        if h in _in_progress_items:
            logger.debug(f"{h} - waiting for processing")
            await _in_progress_items[h].wait()
            logger.debug(f"{h} - done waiting for processing")

        # Is it already done?
        r = sx._cache.lookup_inmem(h)
        if r is not None:
            return r

        # It is not. We need to calculate it, and prevent
        # others from working on it.
        _in_progress_items[h] = asyncio.Event()
        try:
            logger.debug(f"{h} - processing request")
            result = await fn(*args, **kwargs)
            sx._cache.set_inmem(h, result)
        finally:
            logger.debug(f"{h} - done processing request")
            _in_progress_items[h].set()
            del _in_progress_items[h]

        return result

    return cached_version_of_fn



from servicex.ConfigSettings import ConfigSettings
from pathlib import Path

import pytest


@pytest.fixture
def confuse_config():
    config = ConfigSettings("servicex_test_settings", "servicex")
    config.clear()
    config.read(user=False)
    return config


def get_it(c: ConfigSettings):
    return c["testing_value"].get(int)


def test_package_default(confuse_config):
    assert get_it(confuse_config) == 10


def test_local_file_default(confuse_config):
    config_file = Path(".servicex_test_settings")
    try:
        with config_file.open("w") as f:
            f.write("testing_value: 20\n")

        confuse_config.clear()
        confuse_config.read()

        assert get_it(confuse_config) == 20

    finally:
        config_file.unlink()


def test_home_file_default(confuse_config):
    config_file = Path.home() / ".servicex_test_settings"
    try:
        with config_file.open("w") as f:
            f.write("testing_value: 30\n")

        confuse_config.clear()
        confuse_config.read()

        assert get_it(confuse_config) == 30

    finally:
        config_file.unlink()


def test_local_more_important_than_home(confuse_config):
    "Make sure that we pick the local directory over the home one"
    local_config_file = Path(".servicex_test_settings")
    home_config_file = Path.home() / ".servicex_test_settings"
    try:
        with home_config_file.open("w") as f:
            f.write("testing_value: 30\n")
        with local_config_file.open("w") as f:
            f.write("testing_value: 20\n")

        confuse_config.clear()
        confuse_config.read()

        assert get_it(confuse_config) == 20

    finally:
        local_config_file.unlink()
        home_config_file.unlink()


def test_one_level_up(confuse_config):
    "Make sure the config file that is one file up is found"
    config_file = Path(".").resolve().parent / (".servicex_test_settings")
    try:
        with config_file.open("w") as f:
            f.write("testing_value: 30\n")

        confuse_config.clear()
        confuse_config.read()

        assert get_it(confuse_config) == 30

    finally:
        config_file.unlink()


def test_local_more_importnat_than_one_level_up(confuse_config):
    "Assure that our local file is found first"
    one_up_config_file = Path(".").resolve().parent / (".servicex_test_settings")
    config_file = Path(".").resolve() / (".servicex_test_settings")
    try:
        with config_file.open("w") as f:
            f.write("testing_value: 30\n")
        with one_up_config_file.open("w") as f:
            f.write("testing_value: 20\n")

        confuse_config.clear()
        confuse_config.read()

        assert get_it(confuse_config) == 30

    finally:
        config_file.unlink()
        one_up_config_file.unlink()



from json import dumps
from servicex import ServiceXFatalTransformException
from typing import Optional, Union

import aiohttp
import pytest

from servicex import (
    ServiceXAdaptor,
    ServiceXException,
    ServiceXFailedFileTransform,
    ServiceXUnknownRequestID,
)
from servicex.servicex_adaptor import (
    transform_status_stream,
    trap_servicex_failures,
)

from .conftest import ClientSessionMocker, as_async_seq


@pytest.fixture
def servicex_status_request(mocker):
    """
    Fixture that emulates the async python library get call when used with a
    status.

      - Does not check the incoming http address
      - Does not check the Returns a standard triple status from servicex
      - Does not check the headers
      - Call this to set:
            servicex_status_request(1, 2, 3)
            Sets remaining to 1, failed to 2, and processed to 3.
    """
    files_remaining = None
    files_failed = None
    files_processed = 0

    def get_status(a, headers=None):
        r = {}

        def store(name: str, values: Optional[Union[int, str]]):
            nonlocal r
            if values is not None:
                r[name] = values

        store("files-remaining", files_remaining)
        store("files-skipped", files_failed)
        store("files-processed", files_processed)
        store("status", "Running")

        return ClientSessionMocker(dumps(r), 200)

    mocker.patch("aiohttp.ClientSession.get", side_effect=get_status)

    def set_it_up(
        remain: Optional[int], failed: Optional[int], processed: Optional[int]
    ):
        nonlocal files_remaining, files_failed, files_processed
        files_remaining = remain
        files_failed = failed
        files_processed = processed

    return set_it_up


@pytest.fixture
def servicex_status_fatal(mocker):
    "Returns a fatal status"
    result = {
        "files-processed": 0,
        "files-remaining": None,
        "files-skipped": 0,
        "request-id": "24e59fa2-e1d7-4831-8c7e-82b2efc7c658",
        "stats": None,
        "status": "Fatal",
    }
    mocker.patch(
        "aiohttp.ClientSession.get",
        return_value=ClientSessionMocker(dumps(result), 200),
    )


@pytest.fixture
def servicex_status_negative(mocker):
    "Returns a weird negative remaining files bug from sx"
    result = {
        "files-processed": 23,
        "files-remaining": -2,
        "files-skipped": 0,
        "request-id": "24e59fa2-e1d7-4831-8c7e-82b2efc7c658",
        "stats": None,
        "status": "Complete",
    }
    mocker.patch(
        "aiohttp.ClientSession.get",
        return_value=ClientSessionMocker(dumps(result), 200),
    )


@pytest.fixture
def good_submit(mocker):
    client = mocker.MagicMock()
    r = ClientSessionMocker(dumps({"request_id": "111-222-333-444"}), 200)
    client.post = mocker.MagicMock(return_value=r)
    return client


@pytest.fixture
def good_submit_with_auth(mocker):
    client = mocker.MagicMock()
    r = ClientSessionMocker(
        [
            dumps({"access_token": "jwt:access"}),
            dumps({"request_id": "111-222-333-444"}),
        ],
        [200, 200],
    )
    client.post = mocker.MagicMock(return_value=r)
    return client


@pytest.fixture
def bad_submit(mocker):
    client = mocker.MagicMock()
    r = ClientSessionMocker(dumps({"message": "bad text"}), 400)
    client.post = lambda d, json, headers: r
    return client


@pytest.fixture
def bad_submit_html(mocker):
    """
    Instead of returning json, it returns text/html.
    """
    client = mocker.MagicMock(spec=aiohttp.ClientSession)

    class bad_return:
        async def __aexit__(self, exc_type, exc, tb):
            pass

        async def __aenter__(self):
            return self

        async def json(self):
            raise Exception("ContentTypeError")

        def status(self):
            return 500

        async def text(self):
            return "html error content bogus world"

    client.post.return_value = bad_return()
    return client


@pytest.fixture
def good_update(mocker):
    client = mocker.MagicMock()
    r = ClientSessionMocker(dumps({"request_id": "111-222-333"}), 200)
    client.get = mocker.MagicMock(return_value=r)
    return client


@pytest.fixture
def bad_update(mocker):
    client = mocker.MagicMock()
    r = ClientSessionMocker(dumps({"message": "Not known"}), 400)
    client.get = mocker.MagicMock(return_value=r)
    return client


@pytest.fixture
def servicex_status_unknown(mocker):
    r = ClientSessionMocker(dumps({"message": "unknown status"}), 500)
    mocker.patch("aiohttp.ClientSession.get", return_value=r)


@pytest.mark.asyncio
async def test_status_no_auth(servicex_status_request):
    servicex_status_request(None, 0, 10)
    sa = ServiceXAdaptor(endpoint="http://localhost:5000/sx")
    async with aiohttp.ClientSession() as client:
        r = await sa.get_transform_status(client, "123-123-123-444")
        assert len(r) == 3
        assert r[0] is None
        assert r[1] == 10
        assert r[2] == 0


@pytest.mark.asyncio
async def test_status_fatal_status(servicex_status_fatal):

    sa = ServiceXAdaptor("http://localhost:500/sx")
    async with aiohttp.ClientSession() as client:
        with pytest.raises(ServiceXFatalTransformException) as e:
            await sa.get_transform_status(client, "123-123-123-444")

        assert "Fatal" in str(e.value)


@pytest.mark.asyncio
async def test_status_negative_status(servicex_status_negative):

    sa = ServiceXAdaptor("http://localhost:500/sx")
    async with aiohttp.ClientSession() as client:
        with pytest.raises(ServiceXFatalTransformException) as e:
            await sa.get_transform_status(client, "123-123-123-444")

        assert "negative" in str(e.value)


@pytest.mark.asyncio
async def test_status_with_auth(mocker):
    client = mocker.MagicMock()
    client.post = mocker.Mock(
        return_value=ClientSessionMocker([dumps({"access_token": "jwt:access"})], [200])
    )

    client.get = mocker.Mock(
        return_value=ClientSessionMocker(
            [dumps({"files-remaining": 1, "files-skipped": 0, "files-processed": 0})],
            [200],
        )
    )

    sa = ServiceXAdaptor(
        endpoint="http://localhost:5000/sx", refresh_token="jwt:refresh"
    )

    await sa.get_transform_status(client, "123-123-123-444")
    client.post.assert_called_with(
        "http://localhost:5000/sx/token/refresh",
        json=None,
        headers={"Authorization": "Bearer jwt:refresh"},
    )
    client.get.assert_called_with(
        "http://localhost:5000/sx/servicex/transformation/123-123-123-444/status",
        headers={"Authorization": "Bearer jwt:access"},
    )


@pytest.mark.asyncio
async def test_status_unknown_request(servicex_status_unknown):

    sa = ServiceXAdaptor("http://localhost:5000/sx")
    with pytest.raises(ServiceXUnknownRequestID) as e:
        async with aiohttp.ClientSession() as client:
            await sa.get_transform_status(client, "123-123-123-444")

    assert "transform status" in str(e.value)


def version_mock(mocker, spec):
    import sys

    if sys.version_info[1] < 8:
        from asyncmock import AsyncMock  # type: ignore

        return AsyncMock(spec=spec)
    else:
        return mocker.MagicMock(spec=spec)


@pytest.mark.asyncio
async def test_status_stream_simple_sequence(mocker):
    adaptor = version_mock(mocker, spec=ServiceXAdaptor)
    adaptor.get_transform_status.configure_mock(return_value=(0, 1, 1))

    async with aiohttp.ClientSession() as client:
        v = [a async for a in transform_status_stream(adaptor, client, "123-455")]

    assert len(v) == 1
    assert v[0] == (0, 1, 1)


@pytest.mark.asyncio
async def test_status_stream_simple_2sequence(short_status_poll_time, mocker):
    adaptor = version_mock(mocker, spec=ServiceXAdaptor)
    adaptor.get_transform_status.configure_mock(side_effect=[(1, 1, 1), (0, 1, 1)])

    async with aiohttp.ClientSession() as client:
        v = [a async for a in transform_status_stream(adaptor, client, "123-455")]

    assert len(v) == 2
    assert v[0] == (1, 1, 1)
    assert v[1] == (0, 1, 1)


@pytest.mark.asyncio
async def test_watch_no_fail(short_status_poll_time, mocker):
    v = [a async for a in trap_servicex_failures(as_async_seq([(1, 0, 0), (0, 1, 0)]))]

    assert len(v) == 2
    assert v[0] == (1, 0, 0)
    assert v[1] == (0, 1, 0)


@pytest.mark.asyncio
async def test_watch_fail(short_status_poll_time, mocker):
    v = []
    with pytest.raises(ServiceXFailedFileTransform) as e:
        async for a in trap_servicex_failures(as_async_seq([(1, 0, 0), (0, 0, 1)])):
            v.append(a)

    # Should force a failure as soon as it is detected.
    assert len(v) == 1
    assert "failed to transform" in str(e.value)


@pytest.mark.asyncio
async def test_watch_fail_start(short_status_poll_time, mocker):
    v = []
    with pytest.raises(ServiceXFailedFileTransform) as e:
        async for a in trap_servicex_failures(
            as_async_seq([(2, 0, 0), (1, 0, 1), (0, 1, 1)])
        ):
            v.append(a)

    assert len(v) == 1
    assert "failed to transform" in str(e.value)


@pytest.mark.asyncio
async def test_submit_good_no_auth(good_submit):
    sa = ServiceXAdaptor(endpoint="http://localhost:5000/sx")

    rid = await sa.submit_query(good_submit, {"hi": "there"})

    good_submit.post.assert_called_once()
    args, kwargs = good_submit.post.call_args

    assert len(args) == 1
    assert args[0] == "http://localhost:5000/sx/servicex/transformation"

    assert len(kwargs) == 2
    assert "headers" in kwargs
    assert len(kwargs["headers"]) == 0

    assert "json" in kwargs
    assert kwargs["json"] == {"hi": "there"}

    assert rid is not None
    assert isinstance(rid, dict)
    assert "request_id" in rid
    assert rid["request_id"] == "111-222-333-444"


@pytest.mark.asyncio
async def test_submit_good_with_auth(mocker):
    client = mocker.MagicMock()
    client.post = mocker.Mock(
        return_value=ClientSessionMocker(
            [
                dumps({"access_token": "jwt:access"}),
                dumps({"request_id": "111-222-333-444"}),
            ],
            [200, 200],
        )
    )

    sa = ServiceXAdaptor(
        endpoint="http://localhost:5000/sx", refresh_token="jwt:refresh"
    )

    await sa.submit_query(client, {"hi": "there"})
    r = client.post.mock_calls

    assert len(r) == 2

    # Verify the login POST
    _, args, kwargs = r[0]
    assert args[0] == "http://localhost:5000/sx/token/refresh"
    assert kwargs["headers"]["Authorization"] == "Bearer jwt:refresh"

    # Verify the Submit POST
    _, args, kwargs = r[1]
    assert args[0] == "http://localhost:5000/sx/servicex/transformation"
    assert kwargs["headers"]["Authorization"] == "Bearer jwt:access"


@pytest.mark.asyncio
async def test_submit_good_with_auth_existing_token(mocker):
    client = mocker.MagicMock()
    client.post = mocker.Mock(
        return_value=ClientSessionMocker(
            [
                dumps({"access_token": "jwt:access"}),
                dumps({"request_id": "111-222-333-444"}),
                dumps({"request_id": "222-333-444-555"}),
            ],
            [200, 200, 200],
        )
    )

    sa = ServiceXAdaptor(
        endpoint="http://localhost:5000/sx", refresh_token="jwt:refresh"
    )

    mocker.patch(
        "google.auth.jwt.decode", return_value={"exp": float("inf")}
    )  # Never expires
    rid1 = await sa.submit_query(client, {"hi": "there"})
    assert rid1["request_id"] == "111-222-333-444"

    rid2 = await sa.submit_query(client, {"hi": "there"})
    assert rid2["request_id"] == "222-333-444-555"

    r = client.post.mock_calls

    assert len(r) == 3

    # Verify the access token POST
    _, args, kwargs = r[0]
    assert args[0] == "http://localhost:5000/sx/token/refresh"
    assert kwargs["headers"]["Authorization"] == "Bearer jwt:refresh"

    # Verify the Submit POST
    _, args, kwargs = r[1]
    assert args[0] == "http://localhost:5000/sx/servicex/transformation"
    assert kwargs["headers"]["Authorization"] == "Bearer jwt:access"

    # Verify the second Submit POST
    _, args, kwargs = r[2]
    assert args[0] == "http://localhost:5000/sx/servicex/transformation"
    assert kwargs["headers"]["Authorization"] == "Bearer jwt:access"


@pytest.mark.asyncio
async def test_submit_good_with_auth_expired_token(mocker):
    client = mocker.MagicMock()
    client.post = mocker.Mock(
        return_value=ClientSessionMocker(
            [
                dumps({"access_token": "jwt:access"}),
                dumps({"request_id": "111-222-333-444"}),
                dumps({"access_token": "jwt:access2"}),
                dumps({"request_id": "222-333-444-555"}),
            ],
            [200, 200, 200, 200],
        )
    )

    sa = ServiceXAdaptor(
        endpoint="http://localhost:5000/sx", refresh_token="jwt:refresh"
    )

    mocker.patch("google.auth.jwt.decode", return_value={"exp": 0})  # Always expired

    rid1 = await sa.submit_query(client, {"hi": "there"})
    assert rid1["request_id"] == "111-222-333-444"

    rid2 = await sa.submit_query(client, {"hi": "there"})
    assert rid2["request_id"] == "222-333-444-555"

    r = client.post.mock_calls

    assert len(r) == 4

    # Verify the login POST
    _, args, kwargs = r[0]
    assert args[0] == "http://localhost:5000/sx/token/refresh"
    assert kwargs["headers"]["Authorization"] == "Bearer jwt:refresh"

    # Verify the Submit POST
    _, args, kwargs = r[1]
    assert args[0] == "http://localhost:5000/sx/servicex/transformation"
    assert kwargs["headers"]["Authorization"] == "Bearer jwt:access"

    # Verify the second login POST
    _, args, kwargs = r[2]
    assert args[0] == "http://localhost:5000/sx/token/refresh"
    assert kwargs["headers"]["Authorization"] == "Bearer jwt:refresh"

    # Verify the second Submit POST
    _, args, kwargs = r[3]
    assert args[0] == "http://localhost:5000/sx/servicex/transformation"
    assert kwargs["headers"]["Authorization"] == "Bearer jwt:access2"


@pytest.mark.asyncio
async def test_submit_bad(bad_submit):
    sa = ServiceXAdaptor(endpoint="http://localhost:5000/sx")

    with pytest.raises(ServiceXException) as e:
        await sa.submit_query(bad_submit, {"hi": "there"})

    assert "bad text" in str(e.value)


@pytest.mark.asyncio
async def test_submit_bad_html(bad_submit_html):
    sa = ServiceXAdaptor(endpoint="http://localhost:5000/sx")

    with pytest.raises(ServiceXException) as e:
        await sa.submit_query(bad_submit_html, {"hi": "there"})

    assert "html" in str(e.value)


@pytest.mark.asyncio
async def test_submit_good_with_bad_token(mocker):
    client = mocker.MagicMock()
    client.post = mocker.Mock(
        return_value=ClientSessionMocker(dumps({"message": "Wrong credentials"}), 401)
    )

    sa = ServiceXAdaptor(endpoint="http://localhost:5000/sx", refresh_token="XXXXX")

    with pytest.raises(ServiceXException) as e:
        await sa.submit_query(client, {"hi": "there"})

    assert "ServiceX access token request rejected" in str(e.value)


@pytest.mark.asyncio
async def test_update_query_status(good_update):
    sa = ServiceXAdaptor(endpoint="http://localhost:5000/sx")

    rid = await sa.get_query_status(good_update, "111-222-333")

    good_update.get.assert_called_once()
    args, kwargs = good_update.get.call_args

    assert len(args) == 1
    assert args[0] == "http://localhost:5000/sx/servicex/transformation/111-222-333"

    assert len(kwargs) == 1
    assert "headers" in kwargs
    assert len(kwargs["headers"]) == 0

    assert rid is not None
    assert isinstance(rid, dict)
    assert "request_id" in rid
    assert rid["request_id"] == "111-222-333"


@pytest.mark.asyncio
async def test_update_query_status_bad(bad_update):
    sa = ServiceXAdaptor(endpoint="http://localhost:5000/sx")

    with pytest.raises(ServiceXException) as e:
        await sa.get_query_status(bad_update, "111-222-333")

    assert "rejected" in str(e.value)


@pytest.mark.asyncio
async def test_fetch_errors_for_bad_req_id(mocker):
    client = mocker.MagicMock()
    client.get = mocker.Mock(return_value=ClientSessionMocker("Request not found", 404))

    x = ServiceXAdaptor("http://localhost:5000")
    with pytest.raises(ServiceXUnknownRequestID):
        await x.dump_query_errors(client, "111-222-333")


@pytest.mark.asyncio
async def test_fetch_errors_for_bad_bad_bad(mocker):
    client = mocker.MagicMock()
    client.get = mocker.Mock(return_value=ClientSessionMocker("Internal Error", 500))

    x = ServiceXAdaptor("http://localhost:5000")
    with pytest.raises(ServiceXException):
        await x.dump_query_errors(client, "111-222-333")


@pytest.mark.asyncio
async def test_fetch_errors_for_req_good(mocker, caplog):
    client = mocker.MagicMock()
    rtn = {
        "errors": [
            {
                "pod-name": "transformer-8716d52a-974e-4f1d-beed-3425485d8f9b-f9785979892dp4",
                "file": "root://xrootd.echo.stfc.ac.uk:1094/atlas:datadisk/rucio/mc16_13TeV/bf/23/DAOD_STDM3.20425969._000001.pool.root.1",  # NOQA
                "events": 0,
                "info": "error: Failed to transform input file root://xrootd.echo.stfc.ac.uk:1094/atlas:datadisk/rucio/mc16_13TeV/7e/e5/DAOD_STDM3.20425969._000077.pool.root.1: Output\nAnother line",  # NOQA
            }
        ]
    }
    client.get = mocker.Mock(return_value=ClientSessionMocker(dumps(rtn), 200))

    x = ServiceXAdaptor("http://localhost:5000")
    await x.dump_query_errors(client, "111-222-333")
    client.get.assert_called_with(
        "http://localhost:5000/servicex/transformation/111-222-333/errors", headers={}
    )  # NOQA

    warning_messages = (r for r in caplog.records if r.levelname == "WARNING")

    assert sum((1 for w in warning_messages)) > 0
